"""
Embedded Atom Models (EAM)
"""

from __future__ import annotations

from typing import Callable

import torch
from graph_pes.core import GraphPESModel
from graph_pes.data import AtomicGraph
from torch_geometric.utils import scatter

DistanceEmbedding = Callable[[torch.Tensor], torch.Tensor]
"""
embeds a distance tensor into a feature space
"""
Readout = Callable[[torch.Tensor], torch.Tensor]
"""
reads out a single value from a feature tensor
"""


class EAM(GraphPESModel):
    def __init__(
        self,
        embedding: DistanceEmbedding,
        readout: Readout,
        aggregation: str = "sum",
    ):
        super().__init__()
        self.embedding = embedding
        self.readout = readout
        self.aggregation = aggregation

    def predict_local_energies(self, graph: AtomicGraph) -> torch.Tensor:
        # 1. embed the distances
        distance_embeddings = self.embedding(
            graph.neighbour_distances.view(-1, 1)
        )
        # 2. aggregate the embeddings over the neighbours
        central_atoms, neighbours = graph.neighbour_index
        agg_embeddings = scatter(
            distance_embeddings.squeeze(),
            central_atoms,
            dim=0,
            reduce=self.aggregation,
        ).view(graph.n_atoms, -1)
        # 3. read out the local energies
        return self.readout(agg_embeddings)



from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from graph_pes.data import (
    AtomicGraph,
    AtomicGraphBatch,
    GlobalProperty,
    LocalProperty,
    sum_per_structure,
)
from graph_pes.nn import PerSpeciesParameter
from jaxtyping import Shaped
from torch import Tensor, nn


class Transform(nn.Module, ABC):
    r"""
    Abstract base class for shape-preserving transformations of properties,
    :math:`x`, defined on :class:`AtomicGraph <graph_pes.data.AtomicGraph>`s,
    :math:`\mathcal{G}`.

    :math:`T: (x; \mathcal{G}) \mapsto y, \quad x, y \in \mathbb{R}^n`

    Subclasses should implement :meth:`forward`, :meth:`inverse`,
    and :meth:`fit`.

    Parameters
    ----------
    trainable
        Whether the transform should be trainable.
    """

    def __init__(self, trainable: bool = True):
        super().__init__()
        self.trainable = trainable

    @abstractmethod
    def forward(
        self,
        x: LocalProperty | GlobalProperty,
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> LocalProperty | GlobalProperty:
        r"""
        Implements the forward transformation, :math:`y = T(x; \mathcal{G})`.

        Parameters
        ----------
        x
            The input data.
        graph
            The graph to condition the transformation on.

        Returns
        -------
        y: Tensor
            The transformed data.
        """

    @abstractmethod
    def inverse(
        self,
        x: LocalProperty | GlobalProperty,
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> LocalProperty | GlobalProperty:
        r"""
        Implements the inverse transformation,
        :math:`x = T^{-1}(y; \mathcal{G})`.

        Parameters
        ----------
        x
            The input data.
        graph
            The graph to condition the inverse transformation on.

        Returns
        -------
        x: Tensor
            The inversely-transformed data.
        """

    @abstractmethod
    @torch.no_grad()
    def fit(self, x: Tensor, graphs: AtomicGraphBatch) -> Transform:
        r"""
        Fits the transform to property `x` defined on `graphs`.

        Parameters
        ----------
        x
            The property to fit to.
        graphs
            The graphs :math:`\mathcal{G}` that the data originates from.
        """


class Identity(Transform):
    r"""
    The identity transform :math:`T(x; \mathcal{G}) = x` (provided
    for convenience).
    """

    def __init__(self):
        super().__init__(trainable=False)

    def forward(self, x, graph):
        return x

    def inverse(self, x, graph):
        return x

    def fit(self, x: Tensor, graphs: AtomicGraphBatch) -> Transform:
        return self


class Chain(Transform):
    r"""
    A chain of transformations, :math:`T_n \circ \dots \circ T_2 \circ T_1`.

    The forward transformation is applied sequentially from left to right,
    :math:`y = T_n \circ \dots \circ T_2 \circ T_1(x; \mathcal{G})`.

    The inverse transformation is applied sequentially from right to left,
    :math:`x = T_1^{-1} \circ T_2^{-1} \circ \dots
    \circ T_n^{-1}(y; \mathcal{G})`.

    Parameters
    ----------
    transforms
        The transformations to chain together.
    trainable
        Whether the chain should be trainable.
    """

    def __init__(self, transforms: list[Transform], trainable: bool = True):
        super().__init__(trainable)
        for t in transforms:
            t.trainable = trainable
        self.transforms: list[Transform] = nn.ModuleList(transforms)  # type: ignore

    def forward(self, x, graph):
        for transform in self.transforms:
            x = transform(x, graph)
        return x

    def inverse(self, x, graph):
        for transform in reversed(self.transforms):
            x = transform.inverse(x, graph)
        return x

    @torch.no_grad()
    def fit(self, x: Tensor, graphs: AtomicGraphBatch) -> Transform:
        for transform in self.transforms:
            transform.fit(x, graphs)
            x = transform(x, graphs)
        return self


class PerAtomShift(Transform):
    """
    Applies a (species-dependent) per-atom shift to either a local
    or global property.

    Within :meth:`fit`, we calculate the per-species shift that center the
    input data about 0.

    Within :meth:`forward`, we apply the fitted shift to the input data, and
    hence expect the output to be centered about 0.

    Within :meth:`inverse`, we apply the inverse of the fitted shift to the
    input data. If this input is centered about 0, we expect the output to be
    centered about the fitted shift.

    Parameters
    ----------
    trainable
        Whether the shift should be trainable. If ``True``, the fitted
        shift can be changed during training. If ``False``, the fitted
        shift is fixed.
    """

    def __init__(self, trainable: bool = True):
        super().__init__(trainable=trainable)
        self.shift = PerSpeciesParameter.of_dim(
            dim=1, requires_grad=trainable, generator=0
        )
        """The fitted, per-species shifts."""

    @torch.no_grad()
    def fit(self, x: LocalProperty | GlobalProperty, graphs: AtomicGraphBatch):
        r"""
        Fit the shift to the data, :math:`x`.

        Where :math:`x` is a local, per-atom property, we fit the shift
        to be the mean of :math:`x` per species.

        Where :math:`x` is a global property, we assume that this property
        is a sum of local properties, and so perform linear regression to
        get the shift per species.

        Parameters
        ----------
        x
            The input data.
        graphs
            The atomic graphs that x originates from.
        """
        # reset the shift
        zs = torch.unique(graphs.Z)

        if graphs.is_local_property(x):
            # we have one data point per atom in the batch
            # we therefore fit the shift to be the mean of x
            # per unique species
            for z in zs:
                self.shift[z] = x[graphs.Z == z].mean()

        else:
            # we have a single data point per structure in the batch
            # we assume that x is produced as a sum of local properties
            # and do linear regression to guess the shift per species
            N = torch.zeros(graphs.n_structures, len(zs))
            for idx, z in enumerate(zs):
                N[:, idx] = sum_per_structure((graphs.Z == z).float(), graphs)
            shift_vec = torch.linalg.lstsq(N, x).solution
            for idx, z in enumerate(zs):
                self.shift[z] = shift_vec[idx]

    def forward(
        self,
        x: Shaped[Tensor, "shape ..."],
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> Shaped[Tensor, "shape ..."]:
        r"""
        Subtract the learned shift from :math:`x` such that the output
        is expected to be centered about 0 if :math:`x` is centered similarly
        to the data used to fit the shift.

        If :math:`x` is a local property, we subtract the shift from
        each element: :math:`x_i \rightarrow x_i - \text{shift}_i`.

        If :math:`x` is a global property, we subtract the shift from
        each structure: :math:`x_i \rightarrow x_i - \sum_{j \in i}
        \text{shift}_j`.

        Parameters
        ----------
        x
            The input data.
        batch
            The batch of atomic graphs.

        Returns
        -------
        Shaped[Tensor, "shape ..."]
            The input data, shifted by the learned shift.
        """
        shifts = self.shift[graph.Z].squeeze()
        if not graph.is_local_property(x):
            shifts = sum_per_structure(shifts, graph)

        ndims = len(x.shape)
        return x - shifts.view(-1, *([1] * (ndims - 1)))

    def inverse(
        self,
        x: Shaped[Tensor, "shape ..."],
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> Shaped[Tensor, "shape ..."]:
        r"""
        Add the learned shift to :math:`x`, such that the output
        is expected to be centered about the learned shift if :math:`x`
        is centered about 0.

        If :math:`x` is a local property, we add the shift to
        each element: :math:`x_i \rightarrow x_i + \text{shift}_i`.

        If :math:`x` is a global property, we add the shift to
        each structure: :math:`x_i \rightarrow x_i + \sum_{j \in i}
        \text{shift}_j`.

        Parameters
        ----------
        x
            The input data.
        batch
            The batch of atomic graphs.
        """
        shifts = self.shift[graph.Z].squeeze()
        if not graph.is_local_property(x):
            shifts = sum_per_structure(shifts, graph)

        ndims = len(x.shape)
        return x + shifts.view(-1, *([1] * (ndims - 1)))

    def __repr__(self):
        return self.shift.__repr__().replace(
            self.shift.__class__.__name__, self.__class__.__name__
        )


class PerAtomScale(Transform):
    r"""
    Applies a (species-dependent) per-atom scale to either a local
    or global property.

    Within :meth:`fit`, we calculate the per-species scale that
    transforms the input data to have unit variance.

    Within :meth:`forward`, we apply the fitted scale to the input data,
    and hence expect the output to have unit variance.

    Within :meth:`inverse`, we apply the inverse of the fitted scale to the
    input data. If this input has unit variance, we expect the output to
    have variance equal to the fitted scale.

    Parameters
    ----------
    trainable
        Whether the scale should be trainable. If ``True``, the fitted
        scale can be changed during training. If ``False``, the fitted
        scale is fixed.
    """

    def __init__(self, trainable: bool = True, act_on_norms: bool = False):
        super().__init__(trainable=trainable)
        self.scales = PerSpeciesParameter.of_dim(
            dim=1, requires_grad=trainable, generator=1
        )
        """The fitted, per-species scales (variances)."""
        self.act_on_norms = act_on_norms

    @torch.no_grad()
    def fit(self, x: LocalProperty | GlobalProperty, graphs: AtomicGraphBatch):
        r"""
        Fit the scale to the data, :math:`x`.

        Where :math:`x` is a local, per-atom property, we fit the scale
        to be the variance of :math:`x` per species.

        Where :math:`x` is a global property, we assume that this property
        is a sum of local properties, with the variance of this property
        being the sum of variances of the local properties:
        :math:`\sigma^2(x) = \sum_{i \in \text{atoms}} \sigma^2(x_i)`.

        Parameters
        ----------
        x
            The input data.
        graphs
            The atomic graphs that x originates from.
        """
        # reset the scale
        zs = torch.unique(graphs.Z)

        if self.act_on_norms:
            x = x.norm(dim=-1)

        if graphs.is_local_property(x):
            # we have one data point per atom in the batch
            # we therefore fit the scale to be the variance of x
            # per unique species
            for z in zs:
                self.scales[z] = x[graphs.Z == z].var()

        else:
            # TODO: this is very tricky + needs more work
            # for now, we just get a single scale for all species
            scale = (x / graphs.structure_sizes**0.5).var()
            self.scales[zs] = scale

    def forward(
        self,
        x: Shaped[Tensor, "shape ..."],
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> Shaped[Tensor, "shape ..."]:
        r"""
        Scale the input data, :math:`x`, by the learned scale such that the
        output is expected to have unit variance if :math:`x` has unit
        variance.

        If :math:`x` is a local property, we scale each element:
        :math:`x_i \rightarrow x_i / \text{scale}_i`.

        If :math:`x` is a global property, we scale each structure:
        :math:`x_i \rightarrow x_i / \sqrt{\text{scale}_i}`.

        Parameters
        ----------
        x
            The input data.
        batch
            The batch of atomic graphs.

        Returns
        -------
        Shaped[Tensor, "shape ..."]
            The input data, scaled by the learned scale.
        """
        scales = self.scales[graph.Z]
        if not graph.is_local_property(x):
            scales = sum_per_structure(scales, graph)

        ndims = len(x.shape)
        return x / scales.view(-1, *([1] * (ndims - 1))) ** 0.5

    def inverse(
        self,
        x: Shaped[Tensor, "shape ..."],
        graph: AtomicGraph | AtomicGraphBatch,
    ) -> Shaped[Tensor, "shape ..."]:
        r"""
        Scale the input data, :math:`x`, by the inverse of the learned scale
        such that the output is expected to have variance equal to the
        learned scale if :math:`x` has unit variance.

        If :math:`x` is a local property, we scale each element:
        :math:`x_i \rightarrow x_i \times \text{scale}_i`.

        If :math:`x` is a global property, we scale each structure:
        :math:`x_i \rightarrow x_i \times \sqrt{\text{scale}_i}`.

        Parameters
        ----------
        x
            The input data.
        batch
            The batch of atomic graphs.

        Returns
        -------
        Shaped[Tensor, "shape ..."]
            The input data, scaled by the inverse of the learned scale.
        """
        scales = self.scales[graph.Z]
        if not graph.is_local_property(x):
            scales = sum_per_structure(scales, graph)

        ndims = len(x.shape)
        return x * scales.view(-1, *([1] * (ndims - 1))) ** 0.5

    def __repr__(self):
        return self.scales.__repr__().replace(
            self.scales.__class__.__name__, self.__class__.__name__
        )


def PerAtomStandardScaler(trainable: bool = True) -> Transform:
    r"""
    A convenience function for a chain of :class:`PerAtomShift` and
    :class:`PerAtomScale` transforms.
    """
    return Chain([PerAtomShift(trainable), PerAtomScale(trainable)])


class Scale(Transform):
    def __init__(self, trainable: bool = True, scale: float = 1.0):
        super().__init__(trainable=trainable)
        self.scale = nn.Parameter(torch.tensor(scale), requires_grad=trainable)

    def forward(self, x: Tensor, graph: AtomicGraph) -> Tensor:
        return x * self.scale**0.5

    def inverse(self, x: Tensor, graph: AtomicGraph) -> Tensor:
        return x / self.scale**0.5

    @torch.no_grad()
    def fit(self, x: Tensor, graphs: AtomicGraphBatch) -> Transform:
        self.scale.data = x.var()
        return self


class DividePerAtom(Transform):
    """
    A convenience transform for dividing a property by the number of atoms
    in the structure.
    """

    def __init__(self):
        super().__init__(trainable=False)

    def fit(self, x: Tensor, graphs: AtomicGraphBatch) -> Transform:
        return self

    def forward(self, x: Tensor, graph: AtomicGraph) -> Tensor:
        structure_sizes = (
            graph.structure_sizes
            if isinstance(graph, AtomicGraphBatch)
            else graph.n_atoms
        )
        return x / structure_sizes

    def inverse(self, x: Tensor, graph: AtomicGraph) -> Tensor:
        structure_sizes = (
            graph.structure_sizes
            if isinstance(graph, AtomicGraphBatch)
            else graph.n_atoms
        )
        return x * structure_sizes



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import torch
from ase import Atoms
from graph_pes.core import GraphPESModel, energy_and_forces
from graph_pes.data.atomic_graph import AtomicGraph, convert_to_atomic_graph
from graph_pes.data.batching import AtomicGraphBatch
from graph_pes.models.transforms import guess_local_energy_mean_and_std


def parity_plots(
    model: GraphPESModel,
    graphs: list[AtomicGraph],
    axs=None,
    e0=None,
    **kwargs,
):
    model.eval()

    mean, _ = guess_local_energy_mean_and_std(graphs)

    energies, forces = [], []
    pred_energies, pred_forces = [], []
    n_atoms = []

    for graph in graphs:
        try:
            offsets = mean[graph.Z].sum()
            preds = energy_and_forces(model, graph)
            energies.append(graph.labels["energy"].item() - offsets)
            # energies.append(graph.labels["energy"].item())
            forces.append(graph.labels["forces"].numpy())
            pred_energies.append(preds["energy"].detach().item() - offsets)
            # pred_energies.append(preds.energy.detach().item())
            pred_forces.append(preds["forces"].detach().numpy())
            n_atoms.append(graph.n_atoms)
        except Exception:
            pass

    energies = np.array(energies)
    forces = np.vstack(forces)
    pred_energies = np.array(pred_energies)
    pred_forces = np.vstack(pred_forces)
    n_atoms = np.array(n_atoms)
    energies /= n_atoms
    pred_energies /= n_atoms

    if axs is None:
        fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(6, 3))

    # energy parity plot
    if e0 is None:
        e0 = energies.mean()
    axs[0].scatter(energies - e0, pred_energies - e0, s=10, lw=0, **kwargs)
    axs[0].axline((0, 0), slope=1, color="k", ls="--", lw=1)
    axs[0].set_aspect("equal", "datalim")
    axs[0].set_xlabel(r"$E$   (a.u.)")
    axs[0].set_ylabel(r"$\hat{E}$   (a.u.)")

    # force parity plot
    axs[1].scatter(
        forces.flatten(),
        pred_forces.flatten(),
        s=10,
        lw=0,
        alpha=0.2,
        **kwargs,
    )
    axs[1].axline((0, 0), slope=1, color="k", ls="--", lw=1)
    axs[1].set_aspect("equal", "datalim")
    axs[1].set_xlabel(r"$\mathbf{F}$   (eV/Å)")
    axs[1].set_ylabel(r"$\hat{\mathbf{F}}$   (eV/Å)")


def get_dimers(
    dimer: str,
    r_min: float = 0.25,
    r_max: float = 5,
):
    distances = np.linspace(r_min, r_max, 200)

    def _dimer(r):
        return Atoms(dimer, positions=[[0, 0, 0], [0, 0, r]], pbc=False)

    graphs = [
        convert_to_atomic_graph(_dimer(r), cutoff=r_max + 1) for r in distances
    ]

    batch = AtomicGraphBatch.from_graphs(graphs)

    return distances, batch


def dimer_curve(
    model: GraphPESModel,
    dimer: str,
    r_min: float = 0.25,
    r_max: float = 5,
    ax=None,
    **kwargs,
):
    distances, batch = get_dimers(dimer, r_min, r_max)

    with torch.no_grad():
        energies = model(batch).detach().numpy() / 2

    if ax is None:
        ax = plt.gca()

    plot_kwargs = dict(label="$-$".join(dimer))
    plot_kwargs.update(kwargs)
    ax.plot(distances, energies, **plot_kwargs)
    ax.set_xlabel("Distance (Å)")
    ax.set_ylabel("Energy (eV/atom)")



from __future__ import annotations

import warnings
from typing import Sequence

import torch
from jaxtyping import Float, Int, Shaped
from torch import Tensor
from torch.utils.data import DataLoader as TorchDataLoader
from torch_geometric.utils import scatter

from ..util import pairs
from .atomic_graph import AtomicGraph


def sum_per_structure(
    x: torch.Tensor, graph: AtomicGraph | AtomicGraphBatch
) -> torch.Tensor:
    """
    Sum a per-atom property into a per-structure property, respecting
    any batch structure of the input graph.

    Parameters
    ----------
    x
        The per-atom property to sum.
    graph
        The graph to sum over.
    """
    if isinstance(graph, AtomicGraphBatch):
        # we have more than one structure: sum over local energies to
        # get a total energy for each structure
        return scatter(x, graph.batch, dim=0, reduce="sum")
    else:
        # we only have one structure: sum over all the atoms
        return x.sum()


class AtomicDataLoader(TorchDataLoader):
    r"""
    A data loader for merging :class:`AtomicGraph` objects into
    :class:`AtomicGraphBatch` objects.

    Parameters
    ----------
    dataset: Sequence[AtomicGraph]
        The dataset to load.
    batch_size: int
        The batch size.
    shuffle: bool
        Whether to shuffle the dataset.
    **kwargs:
        Additional keyword arguments are passed to the underlying
        :class:`torch.utils.data.DataLoader`.
    """

    def __init__(
        self,
        dataset: Sequence[AtomicGraph],
        batch_size: int = 1,
        shuffle: bool = False,
        **kwargs,
    ):
        if "collate_fn" in kwargs:
            warnings.warn(
                "graph-pes uses a custom collate_fn (`collate_atomic_graphs`), "
                "are you sure you want to override this?",
                stacklevel=2,
            )

        collate_fn = kwargs.pop("collate_fn", _collate_atomic_graphs)

        super().__init__(
            dataset,  # type: ignore
            batch_size,
            shuffle,
            collate_fn=collate_fn,
            **kwargs,
        )


class AtomicGraphBatch(AtomicGraph):
    """
    A disconnected graph representing multiple atomic structures.

    .. note::
        We use `jaxtyping <https://docs.kidger.site/jaxtyping/>`_ notation
        to annotate all tensor shapes. Each AtomicGraphBatch contains
        :math:`B` structures, with a total of :math:`Nb` atoms
        and :math:`Eb` edges.

    Parameters
    ----------
    Z
        The atomic numbers of the atoms.
    positions
        The positions of the atoms.
    neighbour_index
        An edge list specifying neighbouring atoms :math:`j`
        for each central atom :math:`i`.
    cells
        The unit cells of the structure.
    neighbour_offsets
        An edge list specifying the periodic offset of neighbouring atoms
        :math:`j` for each central atom :math:`i`.
    batch
        The structure index of each atom, e.g. `[0, 0, 0, 1, 1, 1, 1, 2, 2, 2]`.
        In this case, the batch contains 3 structures, with 3, 4 and 3 atoms
    ptr
        A pointer to the start and end of each structure in the batch.
        In the above example, this would be `[0, 3, 7, 10]`.
    labels
        Additional, user defined labels for the structure/atoms/edges.
    """

    def __init__(
        self,
        Z: Int[Tensor, "Nb"],
        positions: Float[Tensor, "Nb 3"],
        neighbour_index: Int[Tensor, "2 Eb"],
        cells: Float[Tensor, "B 3 3"],
        neighbour_offsets: Int[Tensor, "Eb 3"],
        batch: Int[Tensor, "Nb"],
        ptr: Int[Tensor, "B+1"],
        atom_labels: dict[str, Shaped[Tensor, "Nb ..."]] | None = None,
        edge_labels: dict[str, Shaped[Tensor, "Eb ..."]] | None = None,
        structure_labels: dict[str, Shaped[Tensor, "B, ..."]] | None = None,
    ):
        super().__init__(
            Z=Z,
            positions=positions,
            neighbour_index=neighbour_index,
            cell=cells,
            neighbour_offsets=neighbour_offsets,
            atom_labels=atom_labels,
            edge_labels=edge_labels,
            structure_labels=structure_labels,
        )

        self.batch = batch
        self.ptr = ptr

    @property
    def neighbour_vectors(self) -> torch.Tensor:
        """
        The vectors from central atoms :math:`i` to neighbours :math:`j`,
        respecting any periodic boundary conditions.
        """

        # more involved than the single structure case,
        # because each cell is different

        i, j = self.neighbour_index

        if not self.has_cell:
            return self._positions[j] - self._positions[i]

        # otherwise calculate offsets on a per-structure basis
        actual_offsets = torch.zeros((self.neighbour_index.shape[1], 3))
        # TODO: parallelise this loop
        for batch, (start, end) in enumerate(pairs(self.ptr)):
            mask = (i >= start) & (i < end)
            actual_offsets[mask] = (
                self.neighbour_offsets[mask].float() @ self.cell[batch]
            )

        return self._positions[j] - self._positions[i] + actual_offsets

    @property
    def n_structures(self) -> int:
        """The number of structures within this batch."""
        return self.ptr.shape[0] - 1

    @property
    def structure_sizes(self) -> torch.Tensor:
        """The number of atoms in each structure within this batch."""
        return self.ptr[1:] - self.ptr[:-1]

    def to(self, device: torch.device | str) -> AtomicGraphBatch:
        def tensor_dict_to_device(
            d: dict[str, torch.Tensor]
        ) -> dict[str, torch.Tensor]:
            return {k: v.to(device) for k, v in d.items()}

        return AtomicGraphBatch(
            Z=self.Z.to(device),
            positions=self._positions.to(device),
            neighbour_index=self.neighbour_index.to(device),
            batch=self.batch.to(device),
            cells=self.cells.to(device),
            neighbour_offsets=self.neighbour_offsets.to(device),
            ptr=self.ptr.to(device),
            structure_labels=tensor_dict_to_device(self.structure_labels),
            atom_labels=tensor_dict_to_device(self.atom_labels),
            edge_labels=tensor_dict_to_device(self.edge_labels),
        )

    @property
    def cells(self) -> torch.Tensor:
        return self.cell

    @classmethod
    def from_graphs(cls, graphs: list[AtomicGraph]) -> AtomicGraphBatch:
        return _collate_atomic_graphs(graphs)


def _collate_atomic_graphs(graphs: list[AtomicGraph]) -> AtomicGraphBatch:
    # easy properties: just cat these together
    Z = torch.cat([g.Z for g in graphs])
    positions = torch.cat([g._positions for g in graphs])
    cells = torch.stack([g.cell for g in graphs])
    neighbour_offsets = torch.cat([g.neighbour_offsets for g in graphs])

    # per-atom and per-edge labels are concatenated along the first axis
    per_atom_labels = {
        k: torch.cat([g.atom_labels[k] for g in graphs])
        for k in graphs[0].atom_labels
    }
    per_edge_labels = {
        k: torch.cat([g.edge_labels[k] for g in graphs])
        for k in graphs[0].edge_labels
    }

    # per-structure labels are concatenated along a new batch axis (0)
    per_structure_labels = {
        k: torch.stack([g.structure_labels[k] for g in graphs])
        for k in graphs[0].structure_labels
    }

    # standard way to calculate the batch and ptr properties
    batch = torch.cat(
        [torch.full((g.Z.shape[0],), i) for i, g in enumerate(graphs)]
    )
    ptr = torch.tensor([0] + [g.Z.shape[0] for g in graphs]).cumsum(dim=0)

    # use the ptr to increment the neighbour index appropriately
    neighbour_index = torch.cat(
        [g.neighbour_index + ptr[i] for i, g in enumerate(graphs)], dim=1
    )

    return AtomicGraphBatch(
        Z=Z,
        positions=positions,
        neighbour_index=neighbour_index,
        cells=cells,
        neighbour_offsets=neighbour_offsets,
        batch=batch,
        ptr=ptr,
        atom_labels=per_atom_labels,
        edge_labels=per_edge_labels,
        structure_labels=per_structure_labels,
    )



from __future__ import annotations

import logging
from pathlib import Path
from typing import Callable

import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint

from .core import GraphPESModel, get_predictions
from .data import AtomicGraph
from .data.batching import AtomicDataLoader, AtomicGraphBatch
from .loss import RMSE, Loss, WeightedLoss
from .transform import Chain, PerAtomScale, Scale
from .util import Keys


def train_model(
    model: GraphPESModel,
    train_data: list[AtomicGraph],
    val_data: list[AtomicGraph] | None = None,
    optimizer: Callable[[], torch.optim.Optimizer] | None = None,
    loss: WeightedLoss | Loss | None = None,
    property_labels: dict[Keys, str] | None = None,
    *,
    batch_size: int = 32,
    pre_fit_model: bool = True,
    # pytorch lightning
    **trainer_kwargs,
):
    # sanity check, but also ensures things like per-atom parameters
    # are registered
    try:
        for graph in train_data[:10]:
            model(graph)
    except Exception as e:
        raise ValueError("The model does not appear to work") from e

    # TODO check using a strict flag that all the data have the same keys
    batch = AtomicGraphBatch.from_graphs(train_data)

    # check that the property keys are valid
    if property_labels is None:
        property_labels = get_existing_keys(batch)
        if not property_labels:
            expected = [key.value for key in Keys.__members__.values()]
            raise ValueError(
                "No property_keys were provided, and none were found in "
                f"the data. Expected at least one of: {expected}"
            )
    else:
        missing = {
            label for label in property_labels.values() if label not in batch
        }
        if missing:
            raise ValueError(
                f"Input data is missing the following keys that were "
                f"requested: {missing}"
            )

    expected_shapes = {
        Keys.ENERGY: (batch.n_structures,),
        Keys.FORCES: (batch.n_atoms, 3),
        Keys.STRESS: (batch.n_structures, 3, 3),
    }
    for key, label in property_labels.items():
        if batch[label].shape != expected_shapes[key]:
            raise ValueError(
                f"Expected {label} to have shape {expected_shapes[key]}, "
                f"but found {batch[label].shape}"
            )
    if Keys.STRESS in property_labels and not batch.has_cell:
        raise ValueError("Can't train on stress without cell information.")

    # create the data loaders
    train_loader = AtomicDataLoader(train_data, batch_size, shuffle=True)
    val_loader = (
        AtomicDataLoader(val_data, batch_size, shuffle=False)
        if val_data is not None
        else None
    )

    # deal with fitting transforms
    # TODO: what if not training on energy?
    if pre_fit_model and Keys.ENERGY in property_labels:
        model.pre_fit(batch, property_labels[Keys.ENERGY])

    actual_loss = get_loss(loss, property_labels)
    actual_loss.fit_transform(batch)

    # deal with the optimizer
    if optimizer is None:
        opt = torch.optim.Adam(model.parameters(), lr=3e-4)
    else:
        opt = optimizer()

    # create the task (a pytorch lightning module)
    task = LearnThePES(model, opt, actual_loss, property_labels)

    # create the trainer
    kwargs = default_trainer_kwargs()
    kwargs.update(trainer_kwargs)
    trainer = pl.Trainer(**kwargs)

    # log info
    params = sum(p.numel() for p in model.parameters())
    device = trainer.accelerator.__class__.__name__.replace("Accelerator", "")
    print(f"Training on : {list(property_labels.values())}")
    print(f"# of params : {params}")
    print(f"Device      : {device}")
    print()

    # train
    trainer.fit(task, train_loader, val_loader)

    # load the best weights
    return task.load_best_weights(model, trainer)


def get_existing_keys(batch: AtomicGraphBatch) -> dict[Keys, str]:
    return {
        key: key.value
        for key in Keys.__members__.values()
        if key.value in batch
    }


class LearnThePES(pl.LightningModule):
    def __init__(
        self,
        model: GraphPESModel,
        optimizer: torch.optim.Optimizer,
        loss: WeightedLoss,
        property_labels: dict[Keys, str],
    ):
        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.loss = loss
        self.property_labels = property_labels

    def forward(self, graphs: AtomicGraphBatch):
        return self.model(graphs)

    def _step(self, graph: AtomicGraphBatch, prefix: str):
        """
        Get (and log) the losses for a training/validation step.
        """

        def log(name, value, verbose=True):
            return self.log(
                f"{prefix}_{name}",
                value,
                prog_bar=verbose and prefix == "val",
                on_step=False,
                on_epoch=True,
                batch_size=graph.n_structures,
            )

        # generate prediction:
        predictions = get_predictions(self.model, graph, self.property_labels)

        # compute the losses
        total_loss = torch.scalar_tensor(0.0, device=self.device)

        for loss, weight in zip(self.loss.losses, self.loss.weights):
            value = loss(predictions, graph)
            # log the unweighted components of the loss
            log(f"{loss.property_key}_{loss.name}", value)
            # but weight them when computing the total loss
            total_loss = total_loss + weight * value

            # log the raw values only during validation
            if prefix == "val":
                raw_value = loss.raw(predictions, graph)
                log(f"{loss.property_key}_raw_{loss.name}", raw_value)

        log("total_loss", total_loss)
        return total_loss

    def training_step(self, structure: AtomicGraphBatch, _):
        return self._step(structure, "train")

    def validation_step(self, structure: AtomicGraphBatch, _):
        return self._step(structure, "val")

    def configure_optimizers(self):
        return self.optimizer

    def on_validation_model_eval(self, *args, **kwargs):
        super().on_validation_model_eval(*args, **kwargs)
        # we override the defaults to turn on gradient tracking for the
        # validation step since we (might) compute the forces using autograd
        torch.set_grad_enabled(True)

    @classmethod
    def load_best_weights(
        cls,
        model: GraphPESModel,
        trainer: pl.Trainer | None = None,
        checkpoint_path: Path | str | None = None,
    ) -> GraphPESModel:
        if checkpoint_path is None and trainer is None:
            raise ValueError(
                "Either trainer or checkpoint_path must be provided"
            )
        if checkpoint_path is None:
            path = trainer.checkpoint_callback.best_model_path  # type: ignore
        else:
            path = Path(checkpoint_path)
        checkpoint = torch.load(path)
        state_dict = {
            k.replace("model.", "", 1): v
            for k, v in checkpoint["state_dict"].items()
            if k.startswith("model.")
        }
        model.load_state_dict(state_dict)
        return model


def get_loss(
    loss: WeightedLoss | Loss | None, property_labels: dict[Keys, str]
) -> WeightedLoss:
    if loss is None:
        default_transforms = {
            Keys.ENERGY: Chain([PerAtomScale(), PerAtomScale()]),
            Keys.FORCES: PerAtomScale(),
            Keys.STRESS: Scale(),
        }
        default_weights = {
            Keys.ENERGY: 1.0,
            Keys.FORCES: 1.0,
            Keys.STRESS: 1.0,
        }
        return WeightedLoss(
            [
                Loss(
                    label,
                    metric=RMSE(),
                    transform=default_transforms[key],
                )
                for key, label in property_labels.items()
            ],
            [default_weights[key] for key in property_labels],
        )
    elif isinstance(loss, Loss):
        return WeightedLoss([loss], [1.0])
    else:
        return loss


def default_trainer_kwargs() -> dict:
    es_callback = EarlyStopping(
        monitor="val_total_loss",
        patience=75,
        mode="min",
        min_delta=1e-3,
    )
    checkpoint_callback = ModelCheckpoint(
        monitor="val_total_loss",
        mode="min",
        save_top_k=1,
        filename="{epoch}-{val_total_loss:.4f}",
    )
    return {
        "accelerator": "auto",
        "max_epochs": 100,
        "enable_model_summary": False,
        "callbacks": [es_callback, checkpoint_callback],
    }


# disable verbose logging from pytorch lightning


def device_info_filter(record):
    return "PU available: " not in record.getMessage()


logging.getLogger("pytorch_lightning.utilities.rank_zero").addFilter(
    device_info_filter
)



from __future__ import annotations

from typing import Callable

import torch
from graph_pes.data import AtomicGraphBatch
from graph_pes.transform import Identity, Transform
from torch import Tensor, nn


class Loss(nn.Module):
    r"""
    Measure the discrepancy between predictions and labels.

    Often, it is convenient to apply some well known loss function,
    e.g. `MSELoss`,  to a transformed version of the predictions and labels,
    e.g. normalisation, such that the loss value takes on "nice" values,
    and that the resulting gradients and parameter updates are well-behaved.

    :class:`Loss`'s in `graph-pes` are thus lightweight wrappers around:

    * an (optional) pre-transform, :math:`T`
    * a loss metric, :math:`M`.

    Calculating the loss value, :math:`\mathcal{L}`, amounts to the following:

    .. math::
        \mathcal{L} = M\left( T(\hat{P}), T(P) \right)

    where :math:`\hat{P}` are the predictions, and :math:`P` are the labels.

    You can create a weighted, multi-component loss by using the `+` and `*`
    operators on :class:`Loss` instances. For example:

    .. code-block:: python

            Loss("energy") * 10 + Loss("forces")

    will create a loss function that is the sum of a 10x weighted energy loss
    and a force loss.

    Parameters
    ----------
    property_key
        The name of the property to apply transformations and loss metrics to.
    metric
        The loss metric to use.
    transform
        The transform to apply to the predictions and labels.
    """

    def __init__(
        self,
        property_key: str,
        metric: Callable[[Tensor, Tensor], Tensor] | None = None,
        transform: Transform | None = None,
    ):
        super().__init__()
        self.property_key = property_key
        self.metric = MAE() if metric is None else metric
        self.transform = transform or Identity()
        self.transform.trainable = False

    def forward(
        self,
        predictions: dict[str, torch.Tensor],
        graphs: AtomicGraphBatch,
    ) -> torch.Tensor:
        """
        Computes the loss value.

        Parameters
        ----------
        predictions
            The predictions from the model.
        graphs
            The graphs containing the labels.
        """

        return self.metric(
            self.transform(predictions[self.property_key], graphs),
            self.transform(graphs[self.property_key], graphs),
        )

    def raw(
        self,
        predictions: dict[str, torch.Tensor],
        graphs: AtomicGraphBatch,
    ) -> torch.Tensor:
        """
        Compute the metric as applied directly to the predictions and labels.

        Parameters
        ----------
        predictions
            The predictions from the model.
        graphs
            The graphs containing the labels.
        """

        return self.metric(
            predictions[self.property_key],
            graphs[self.property_key],
        )

    def fit_transform(self, graphs: AtomicGraphBatch):
        """
        Fit the transform to the labels.

        Parameters
        ----------
        graphs
            The graphs containing the labels.
        """

        self.transform.fit(graphs[self.property_key], graphs)

    @property
    def name(self) -> str:
        # if metric is a class, we want the class name otherwise we want
        # the function name, all without the word "loss" in it
        return (
            getattr(
                self.metric,
                "__name__",
                self.metric.__class__.__name__,
            )
            .lower()
            .replace("loss", "")
        )

    def __mul__(self, other: float) -> WeightedLoss:
        return WeightedLoss([self], [other])

    def __rmul__(self, other: float) -> WeightedLoss:
        return WeightedLoss([self], [other])

    def __add__(self, other: Loss | WeightedLoss) -> WeightedLoss:
        if isinstance(other, Loss):
            return WeightedLoss([self, other], [1, 1])
        elif isinstance(other, WeightedLoss):
            return WeightedLoss([self] + other.losses, [1] + other.weights)
        else:
            raise TypeError(f"Cannot add Loss and {type(other)}")

    def __radd__(self, other: Loss | WeightedLoss) -> WeightedLoss:
        return self.__add__(other)


class WeightedLoss(torch.nn.Module):
    r"""
    A lightweight wrapper around a collection of weighted losses.

    Creation can be done in two ways:

    1. Directly, by passing a list of losses and weights.
    2. Via `+` and `*` operators

    Hence:

    .. code-block:: python

        WeightedLoss([Loss("energy"), Loss("forces")], [10, 1])
        # is equivalent to
        10 * Loss("energy") + 1 * Loss("forces")

    Parameters
    ----------
    losses
        The collection of losses to aggregate.
    weights
        The weights to apply to each loss.
    """

    def __init__(
        self,
        losses: list[Loss],
        weights: list[float] | None = None,
    ):
        super().__init__()
        self.losses: list[Loss] = nn.ModuleList(losses)  # type: ignore
        self.weights = weights or [1.0] * len(losses)

    def __add__(self, other: WeightedLoss) -> WeightedLoss:
        return WeightedLoss(
            self.losses + other.losses, self.weights + other.weights
        )

    def fit_transform(self, graphs: AtomicGraphBatch):
        for loss in self.losses:
            loss.fit_transform(graphs)


class RMSE(torch.nn.MSELoss):
    r"""
    Root mean squared error metric:

    .. math::
        \sqrt{ \frac{1}{N} \sum_i^N \left( \hat{P}_i - P_i \right)^2 }
    """

    def __init__(self):
        super().__init__()

    def forward(
        self, input: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        return (super().forward(input, target)).sqrt()


class MAE(torch.nn.L1Loss):
    r"""
    Mean absolute error metric:

    .. math::
        \frac{1}{N} \sum_i^N \left| \hat{P}_i - P_i \right|
    """

    def __init__(self):
        super().__init__()



from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from graph_pes.core import EnergySummation, GraphPESModel
from graph_pes.data import AtomicGraph
from graph_pes.data.batching import AtomicGraphBatch
from graph_pes.nn import PositiveParameter
from graph_pes.transform import PerAtomScale, PerAtomShift
from jaxtyping import Float
from torch import Tensor, nn

from .distances import Envelope


class PairPotential(GraphPESModel, ABC):
    r"""
    An abstract base class for PES models that calculate system energy as
    a sum over pairwise interactions:

    .. math::
        E = \sum_{i, j} V(r_{ij}, Z_i, Z_j)

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    and :math:`Z_i` and :math:`Z_j` are their atomic numbers.
    This can be recast as a sum over local energy contributions,
    :math:`E = \sum_i \varepsilon_i`, according to:

    .. math::
        \varepsilon_i = \frac{1}{2} \sum_j V(r_{ij}, Z_i, Z_j)

    Subclasses should implement :meth:`interaction`.
    """

    @abstractmethod
    def interaction(
        self,
        r: Float[Tensor, "E"],
        Z_i: Float[Tensor, "E"],
        Z_j: Float[Tensor, "E"],
    ) -> Float[Tensor, "E"]:
        """
        Compute the interactions between pairs of atoms, given their
        distances and atomic numbers.

        Parameters
        ----------
        r
            The pair-wise distances between the atoms.
        Z_i
            The atomic numbers of the central atoms.
        Z_j
            The atomic numbers of the neighbours.

        Returns
        -------
        V: Float[Tensor, "E"]
            The pair-wise interactions.
        """

    def predict_local_energies(
        self, graph: AtomicGraph
    ) -> Float[Tensor, "graph.n_edges"]:
        central_atoms, neighbours = graph.neighbour_index
        distances = graph.neighbour_distances

        Z_i, Z_j = graph.Z[central_atoms], graph.Z[neighbours]
        V = self.interaction(
            distances.view(-1, 1), Z_i.view(-1, 1), Z_j.view(-1, 1)
        )

        # sum over the neighbours
        energies = torch.zeros_like(graph.Z, dtype=torch.float)
        energies.scatter_add_(0, central_atoms, V.squeeze())

        # divide by 2 to avoid double counting
        return energies / 2


class LennardJones(PairPotential):
    r"""
    A pair potential of the form:

    .. math::
        V(r_{ij}, Z_i, Z_j) = V(r_{ij}) = 4 \varepsilon \left[ \left(
        \frac{\sigma}{r_{ij}} \right)^{12} - \left( \frac{\sigma}{r_{ij}}
        \right)^{6} \right]

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`.
    Internally, :math:`\varepsilon` and :math:`\sigma` are stored as
    :class:`PositiveParameter <graph_pes.nn,PositiveParamerer>` instances,
    which ensures that they are kept strictly positive during training.

    Attributes
    ----------
    epsilon: :class:`PositiveParameter <graph_pes.nn.PositiveParameter>`
        The depth of the potential.
    sigma: :class:`PositiveParameter <graph_pes.nn.PositiveParameter>`
        The distance at which the potential is zero.
    """

    def __init__(self):
        super().__init__()
        self.epsilon = PositiveParameter(0.1)
        self.sigma = PositiveParameter(1.0)

        # epsilon is a scaling term, so only need to learn a shift
        # parameter (rather than a shift and scale)
        self._energy_summation = EnergySummation(local_transform=PerAtomShift())

    def interaction(
        self, r: torch.Tensor, Z_i: torch.Tensor, Z_j: torch.Tensor
    ):
        """
        Evaluate the pair potential.

        Parameters
        ----------
        r : torch.Tensor
            The pair-wise distances between the atoms.
        Z_i : torch.Tensor
            The atomic numbers of the central atoms. (unused)
        Z_j : torch.Tensor
            The atomic numbers of the neighbours. (unused)
        """
        x = self.sigma / r
        return 4 * self.epsilon * (x**12 - x**6)

    def pre_fit(self, graph: AtomicGraphBatch, energy_label: str = "energy"):
        super().pre_fit(graph, energy_label)

        # set the distance at which the potential is zero to be
        # close to the minimum pair-wise distance
        d = torch.quantile(graph.neighbour_distances, 0.01)
        self.sigma = PositiveParameter(d)


class Morse(PairPotential):
    r"""
    A pair potential of the form:

    .. math::
        V(r_{ij}, Z_i, Z_j) = V(r_{ij}) = D (1 - e^{-a(r_{ij} - r_0)})^2

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    and :math:`D`, :math:`a` and :math:`r_0` control the depth, width and
    center of the potential well, respectively. Internally, these are stored
    as :class:`PositiveParameter` instances.

    Attributes
    ----------
    D: :class:`PositiveParameter <graph_pes.nn.PositiveParameter>`
        The depth of the potential.
    a: :class:`PositiveParameter <graph_pes.nn.PositiveParameter>`
        The width of the potential.
    r0: :class:`PositiveParameter <graph_pes.nn.PositiveParameter>`
        The center of the potential.
    """

    def __init__(self):
        super().__init__()
        self.D = PositiveParameter(0.1)
        self.a = PositiveParameter(1.0)
        self.r0 = PositiveParameter(0.5)

        # D is a scaling term, so only need to learn a shift
        # parameter (rather than a shift and scale)
        self._energy_summation = EnergySummation(local_transform=PerAtomScale())

    def interaction(
        self, r: torch.Tensor, Z_i: torch.Tensor, Z_j: torch.Tensor
    ):
        """
        Evaluate the pair potential.

        Parameters
        ----------
        r : torch.Tensor
            The pair-wise distances between the atoms.
        Z_i : torch.Tensor
            The atomic numbers of the central atoms. (unused)
        Z_j : torch.Tensor
            The atomic numbers of the neighbours. (unused)
        """
        return self.D * (1 - torch.exp(-self.a * (r - self.r0))) ** 2

    def pre_fit(self, graph: AtomicGraphBatch):
        super().pre_fit(graph)

        # set the potential depth to be shallow
        self.D = PositiveParameter(0.1)

        # set the center of the well to be close to the minimum pair-wise
        # distance
        d = torch.quantile(graph.neighbour_distances, 0.01)
        self.r0 = PositiveParameter(d)

        # set the width to be broad
        self.a = PositiveParameter(0.5)


class LearnablePairPotential(PairPotential):
    r"""
    A pair potential of the form:

    .. math::
        V(r_{ij}, Z_i, Z_j) = V(r_{ij}) = \text{Envelope}(r_{ij}) \circ
        f_\theta(r_{ij})

    where:

    * :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`
    * :math:`f_\theta: \mathbb{R}^{\text{batch} \times 1} \rightarrow
      \mathbb{R}^{\text{batch} \times 1}, r \mapsto V(r)` is a learnable
      function parameterised by :math:`\theta`.
    * :math:`\text{Envelope} : \mathbb{R}^{\text{batch} \times 1} \rightarrow
      \mathbb{R}^{\text{batch} \times 1}` is an envelope function that ensures
      smoothness of the potential at the cutoff distance.

    Parameters
    ----------
    f
        The learnable element.
    envelope
        The envelope.
    """

    def __init__(self, f: nn.Module, envelope: Envelope):
        super().__init__()
        self.f = f
        self.envelope = envelope

    def interaction(
        self,
        r: Float[Tensor, "E 1"],
        Z_i: torch.Tensor | None = None,
        Z_j: torch.Tensor | None = None,
    ) -> Float[Tensor, "E 1"]:
        return self.f(r) * self.envelope(r)



from __future__ import annotations

from typing import Sequence, TypeVar

import numpy as np

__all__ = ["random_split"]

E = TypeVar("E")


def random_split(
    sequence: Sequence[E], lengths: Sequence[int], seed: int | None = None
) -> list[list[E]]:
    """
    Randomly split `sequence` into sub-sequences according to `lengths`.

    Parameters
    ----------
    sequence: Sequence[E]
        The sequence to split.
    lengths: Sequence[int]
        The lengths of the sub-sequences to create.
    seed: int | None
        The random seed to use. If `None`, the current random state is
        used (non-deterministic).

    Returns
    -------
    list[list[E]]
        A list of sub-sequences.

    Examples
    --------
    >>> random_split("abcde", [2, 3])
    [['b', 'c'], ['a', 'd', 'e']]
    """

    if sum(lengths) > len(sequence):
        raise ValueError("Not enough things to split")

    shuffle = np.random.RandomState(seed=seed).permutation(len(sequence))
    ptr = [0, *np.cumsum(lengths)]

    return [
        [sequence[i] for i in shuffle[ptr[n] : ptr[n + 1]]]
        for n in range(len(lengths))
    ]



import torch
from graph_pes.loss import MAE, RMSE, Loss, WeightedLoss


def test_metrics():
    a = torch.tensor([1.0, 2.0, 3.0])
    b = torch.tensor([1.0, 2.0, 3.0])

    assert torch.allclose(MAE()(a, b), torch.tensor(0.0))
    assert torch.allclose(RMSE()(a, b), torch.tensor(0.0))

    c = torch.tensor([0, 0, 0]).float()
    assert torch.allclose(MAE()(a, c), torch.tensor(2.0))
    assert torch.allclose(RMSE()(a, c), torch.tensor((1 + 4 + 9) / 3).sqrt())


def test_loss_ops():
    l1 = Loss("energy")
    l2 = Loss("forces")

    weighted = 10 * l1 + l2 * 1
    assert isinstance(weighted, WeightedLoss)
    assert set(weighted.losses) == {l1, l2}
    assert set(weighted.weights) == {10, 1}



from ase.io import read
from graph_pes.core import get_predictions
from graph_pes.data.atomic_graph import convert_to_atomic_graphs
from graph_pes.data.batching import AtomicGraphBatch
from graph_pes.models.pairwise import LennardJones
from graph_pes.training import Loss, train_model
from graph_pes.util import Keys


def test_integration():
    structures = read("tests/test.xyz", ":")
    graphs = convert_to_atomic_graphs(structures, cutoff=3)
    _batch = AtomicGraphBatch.from_graphs(graphs)

    model = LennardJones()

    loss = Loss("energy")
    property_labels = {Keys.ENERGY: "energy"}
    before = loss(
        get_predictions(model, _batch, property_labels),
        _batch,
    )

    train_model(
        model,
        train_data=graphs[:8],
        val_data=graphs[8:],
        property_labels=property_labels,
        loss=loss,
        max_epochs=2,
        accelerator="cpu",
        callbacks=[],
    )

    after = loss(
        get_predictions(model, _batch, property_labels),
        _batch,
    )

    assert after < before, "training did not improve the loss"



from __future__ import annotations

from typing import Union

from jaxtyping import Shaped
from torch import Tensor
from typing_extensions import TypeAlias

from .atomic_graph import (
    AtomicGraph,
    convert_to_atomic_graph,
    convert_to_atomic_graphs,
)
from .batching import AtomicDataLoader, AtomicGraphBatch, sum_per_structure
from .utils import random_split

LocalProperty: TypeAlias = Shaped[Tensor, "N ..."]
"""
Type alias for a local property, i.e. a property that is defined for each atom.
"""
GlobalProperty: TypeAlias = Union[Tensor, Shaped[Tensor, "B ..."]]
"""
Type alias for a global property, i.e. a property that is defined 
for each structure.
"""

__all__ = [
    "AtomicDataLoader",
    "AtomicGraph",
    "AtomicGraphBatch",
    "convert_to_atomic_graph",
    "convert_to_atomic_graphs",
    "random_split",
    "sum_per_structure",
    "GlobalProperty",
    "LocalProperty",
]



from __future__ import annotations

from typing import Literal

import torch
from graph_pes.core import GraphPESModel
from graph_pes.data.atomic_graph import AtomicGraph
from graph_pes.nn import MLP, PerSpeciesEmbedding, ShiftedSoftplus
from torch import nn
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import scatter

from .distances import (
    Bessel,
    DistanceExpansion,
    GaussianSmearing,
    PolynomialEnvelope,
)


class CFConv(MessagePassing):
    r"""
    CFConv: The Continous Filter Convolution

    Originally proposed as part of `SchNet <https://arxiv.org/abs/1706.08566>`_,
    this message passing layer creates messages of the form:

    .. math::
        m_{ij} = x_j \odot \mathbb{F}(r_{ij})

    where :math:`\mathbb{F}` is some learnable filter function which
    expands the radial distance :math:`r_{ij}` into a set of features
    that match the dimensionality of the node features :math:`x_j`.

    Parameters
    ----------
    filter_generator : nn.Module
        The filter function :math:`\mathbb{F}`.
    """

    def __init__(self, filter_generator: nn.Module):
        super().__init__(aggr="add")
        self.filter_generator = filter_generator

    def message(
        self, x_j: torch.Tensor, neighbour_distances: torch.Tensor
    ) -> torch.Tensor:
        """
        Apply the filter function to the distances and multiply by the
        node features.

        Parameters
        ----------
        x_j : torch.Tensor
            The node features of the neighbors.
        neighbour_distances : torch.Tensor
            The distances to the neighbors.
        """
        filters = self.filter_generator(neighbour_distances)
        return x_j * filters

    def forward(
        self,
        neighbour_index: torch.Tensor,
        neighbour_distances: torch.Tensor,
        node_features: torch.Tensor,
    ) -> torch.Tensor:
        return self.propagate(
            neighbour_index,
            neighbour_distances=neighbour_distances,
            x=node_features,
        )

    def __repr__(self):
        rep = f"CFConv({self.filter_generator})"
        if len(rep) > 40:
            indented = "  " + "\n  ".join(
                str(self.filter_generator).split("\n")
            )
            rep = f"CFConv(\n{indented}\n)"
        return rep


class SchNetInteraction(nn.Module):
    r"""
    SchNet interaction block.

    Updates the embedding of each atom, :math:`x_i` by sequentially
    applying the following:
    - a linear transform of each node's features :math:`h_i \leftarrow W x_i`
    - a message passing cfconv block that replaces the current node's features
        with a sum over the distance-filtered features of its neighbors
        :math:`h_i \leftarrow \sum_{j \in \mathcal{N}(i)}
        \mathbb{F}(r_{ij}) \odot h_j`
    - a multilayer perceptron that further embeds these new node features
        :math:`h_i \leftarrow \mathrm{MLP}(h_i)`

    Parameters
    ----------
    n_features : int
        Number of features per node.
    expansion_features : int
        Number of features used for the radial basis expansion.
    cutoff : float
        Neighborhood cutoff radius.
    """

    def __init__(
        self,
        n_features: int,
        expansion_features: int,
        cutoff: float,
        basis_type: type[DistanceExpansion] = GaussianSmearing,
    ):
        super().__init__()

        # schnet interaction block's are composed of 3 elements

        # 1. linear transform to get new node features
        self.linear = nn.Linear(n_features, n_features)

        # 2. cfconv to mix these new features with distances information,
        # and aggregate over neighbors to create completely new node features
        filter_generator = nn.Sequential(
            basis_type(expansion_features, cutoff),
            MLP(
                [expansion_features, n_features, n_features],
                activation=ShiftedSoftplus(),
            ),
        )
        self.cfconv = CFConv(filter_generator)

        # 3. mlp to further embed these new node features
        self.mlp = MLP(
            [n_features, n_features, n_features],
            activation=ShiftedSoftplus(),
        )

    def forward(
        self,
        neighbour_index: torch.Tensor,
        neighbour_distances: torch.Tensor,
        node_features: torch.Tensor,
    ):
        # 1. linear transform to get new node features
        h = self.linear(node_features)
        # 2. cfconv to mix these new features with distances information,
        # and aggregate over neighbors to create completely new node features
        h = self.cfconv(neighbour_index, neighbour_distances, h)
        # 3. mlp to further embed these new node features
        return self.mlp(h)


class SchNet(GraphPESModel):
    def __init__(
        self,
        node_feature_size: int,
        expansion_feature_size: int,
        cutoff: float,
        num_interactions: int = 3,
        basis_type: type[DistanceExpansion] = GaussianSmearing,
    ):
        super().__init__()

        self.chemical_embedding = PerSpeciesEmbedding(
            dim=node_feature_size, default=None
        )

        self.interactions = nn.ModuleList(
            SchNetInteraction(
                node_feature_size, expansion_feature_size, cutoff, basis_type
            )
            for _ in range(num_interactions)
        )

        self.read_out = MLP(
            [node_feature_size, node_feature_size // 2, 1],
            activation=ShiftedSoftplus(),
        )

    def predict_local_energies(self, graph: AtomicGraph) -> torch.Tensor:
        h = self.chemical_embedding(graph.Z)

        for interaction in self.interactions:
            h = h + interaction(
                graph.neighbour_index, graph.neighbour_distances, h
            )

        return self.read_out(h)


class ImprovedSchNetInteraction(nn.Module):
    def __init__(
        self,
        cutoff: float,
        radial_features: int = 8,
        radial_expansion: type[DistanceExpansion] = Bessel,
        atom_features: int = 4,
        mlp_height: int = 16,
        mlp_depth: int = 3,
        activation: type[nn.Module] = nn.CELU,
        pre_embed: bool = True,
    ):
        super().__init__()

        # radial filter components
        self.envelope = PolynomialEnvelope(cutoff)
        self.raw_filter_generator = nn.Sequential(
            radial_expansion(radial_features, cutoff),
            MLP(
                [radial_features, mlp_height, mlp_height, atom_features],
                activation=activation(),
            ),
        )

        # atom embedding components
        self.pre_embed = pre_embed
        if pre_embed:
            self.pre_embedding = MLP(
                [atom_features] + [mlp_height] * mlp_depth + [atom_features],
                activation=activation(),
            )
        self.post_embedding = MLP(
            [atom_features] + [mlp_height] * mlp_depth + [atom_features],
            activation=activation(),
        )

    def get_radial_embeddings(self, distances: torch.Tensor) -> torch.Tensor:
        envelope = self.envelope(distances).unsqueeze(-1)
        radial_embedding = self.raw_filter_generator(distances)
        return radial_embedding * envelope

    def forward(
        self,
        central_atom_embeddings: torch.Tensor,
        graph: AtomicGraph,
    ) -> torch.Tensor:
        # get radial features
        radial_embedding = self.get_radial_embeddings(graph.neighbour_distances)

        if self.pre_embed:
            # update atomic features
            central_atom_embeddings = self.pre_embedding(
                central_atom_embeddings
            )

        # haddamard product to mix atomic features with radial features
        i, j = graph.neighbour_index
        neighbour_atom_embeddings = central_atom_embeddings[j]
        messages = neighbour_atom_embeddings * radial_embedding

        # aggregate over neighbours to get new central atom embeddings
        central_atom_embeddings = scatter(messages, i, reduce="add")

        # update atomic features
        return self.post_embedding(central_atom_embeddings)


class ImprovedSchNet(GraphPESModel):
    def __init__(
        self,
        cutoff: float,
        radial_features: int = 8,
        radial_expansion: type[DistanceExpansion] = Bessel,
        atom_features: int = 4,
        mlp_height: int = 16,
        mlp_depth: int = 3,
        activation: type[nn.Module] = nn.CELU,
        num_blocks: int = 3,
        residual_messages: bool = True,
        pre_embed: Literal["none", "all", "2+"] = "2+",
    ):
        super().__init__()

        self.residual_messages = residual_messages

        self.chemical_embedding = PerSpeciesEmbedding(dim=atom_features)

        def _rule(i):
            return {
                "none": False,
                "all": True,
                "2+": i > 0,
            }[pre_embed]

        self.interactions = nn.ModuleList(
            ImprovedSchNetInteraction(
                cutoff,
                radial_features,
                radial_expansion,
                atom_features,
                mlp_height,
                mlp_depth,
                activation,
                pre_embed=_rule(i),
            )
            for i in range(num_blocks)
        )

        self.read_outs = nn.ModuleList(
            MLP(
                [atom_features] + [mlp_height] * mlp_depth + [1],
                activation=activation(),
            )
            for _ in range(num_blocks)
        )

    def predict_local_energies(self, graph: AtomicGraph) -> torch.Tensor:
        return self.per_block_predictions(graph).sum(dim=1)

    def per_block_predictions(self, graph: AtomicGraph) -> torch.Tensor:
        h = self.chemical_embedding(graph.Z)

        predictions = []
        for block, read_out in zip(self.interactions, self.read_outs):
            if self.residual_messages:
                h = block(h, graph) + h
            else:
                h = block(h, graph)
            predictions.append(read_out(h).squeeze())

        return torch.stack(predictions, dim=1)



from __future__ import annotations

import warnings
from typing import Iterable

import ase
import numpy as np
import torch
from ase.neighborlist import neighbor_list
from jaxtyping import Float, Int, Shaped
from torch import Tensor

from ..util import as_possible_tensor, shape_repr


class AtomicGraph:
    """
    A graph representation of an atomic structure, compatible with
    PyTorch/Geometric/Lightning.

    .. note::
        We use `jaxtyping <https://docs.kidger.site/jaxtyping/>`_ notation
        to annotate all tensor shapes. Each :class:`AtomicGraph` contains
        :math:`N` atoms and :math:`E` edges

    Parameters
    ----------
    Z
        The atomic numbers of the atoms.
    positions
        The positions of the atoms.
    neighbour_index
        An edge list specifying neighbouring atoms :math:`j`
        for each central atom :math:`i`.
    cell
        The unit cell of the structure.
    neighbour_offsets
        An edge list specifying the periodic offset of neighbouring atoms
        :math:`j` for each central atom :math:`i`.
    atom_labels
        Additional, user-defined, per-atom labels.
    edge_labels
        Additional, user-defined, per-edge labels.
    structure_labels
        Additional, user-defined, per-structure labels.
    """

    def __init__(
        self,
        Z: Int[Tensor, "N"],
        positions: Float[Tensor, "N 3"],
        neighbour_index: Int[Tensor, "2 E"],
        cell: Float[Tensor, "3 3"],
        neighbour_offsets: Int[Tensor, "E 3"],
        atom_labels: dict[str, Shaped[Tensor, "N ..."]] | None = None,
        edge_labels: dict[str, Shaped[Tensor, "E ..."]] | None = None,
        structure_labels: dict[str, Tensor] | None = None,
    ):
        self.Z = Z
        self.neighbour_index = neighbour_index
        self.cell = cell
        self.neighbour_offsets = neighbour_offsets
        self.atom_labels = atom_labels or {}
        self.edge_labels = edge_labels or {}
        self.structure_labels = structure_labels or {}

        # we store positions privately, such that we can warn users
        # who access them directly, via the .position property below,
        # in case they naïvely use them to calculate neighbour vectors
        # or distances, which will be incorrect for periodic systems.
        self._positions = positions

    @classmethod
    def from_isolated_structure(
        cls,
        Z: Int[Tensor, "N"],
        positions: Float[Tensor, "N 3"],
        neighbour_index: Int[Tensor, "2 E"],
        atom_labels: dict[str, Shaped[Tensor, "N ..."]] | None = None,
        edge_labels: dict[str, Shaped[Tensor, "E ..."]] | None = None,
        structure_labels: dict[str, Tensor] | None = None,
    ) -> AtomicGraph:
        """
        Create a graph from a structure with no periodic boundary conditions.

        Parameters
        ----------
        Z
            The atomic numbers of the atoms.
        positions
            The positions of the atoms.
        neighbour_index
            An edge list specifying neighbouring atoms :math:`j`
            for each central atom :math:`i`.
        atom_labels
            Additional, user-defined, per-atom labels.
        edge_labels
            Additional, user-defined, per-edge labels.
        structure_labels
            Additional, user-defined, per-structure labels.
        """
        cell = torch.FloatTensor(torch.zeros((3, 3)))
        neighbour_offsets = torch.Tensor(
            torch.zeros((neighbour_index.shape[1], 3))
        )

        return cls(
            Z=Z,
            positions=positions,
            neighbour_index=neighbour_index,
            cell=cell,
            neighbour_offsets=neighbour_offsets,
            atom_labels=atom_labels,
            edge_labels=edge_labels,
            structure_labels=structure_labels,
        )

    @property
    def has_cell(self) -> bool:
        """Whether the structure has a unit cell."""
        return not torch.all(self.cell == 0).item()

    @property
    def positions(self):
        if self.has_cell:
            # raise a warning if the user accesses the positions directly,
            # since the most likely cause for this is that they are attempting
            # to calculate neighbour vectors/distances as:
            #   neighbour_vectors = positions[j] - positions[i]
            # which will be incorrect for periodic systems.

            warnings.warn(
                "Are you using `AtomicGraph.positions` to calculate "
                "neighbour vectors/distances?\n"
                "Use `AtomicGraph.neighbour_vectors` and "
                "`AtomicGraph.neighbour_distances` instead: for periodic "
                "systems, neighbour_vectors != positions[j] - positions[i]!",
                stacklevel=2,
            )
            # TODO link to docs

        return self._positions

    @property
    def neighbour_vectors(self) -> Float[Tensor, "E 3"]:
        """
        The vectors from central atoms :math:`i` to neighbours :math:`j`,
        respecting any periodic boundary conditions.
        """
        i, j = self.neighbour_index
        if not self.has_cell:
            return self._positions[j] - self._positions[i]

        neighbour_positions = (
            self._positions[j] + self.neighbour_offsets.float() @ self.cell
        )
        return neighbour_positions - self._positions[i]

    @property
    def neighbour_distances(self) -> Float[Tensor, "E"]:
        """
        The distances from central atoms :math:`i` to neighbours :math:`j`,
        respecting any periodic boundary conditions.
        """
        return self.neighbour_vectors.norm(dim=-1)

    @property
    def n_atoms(self) -> int:
        """The number of atoms in the structure."""
        return self.Z.shape[0]

    @property
    def n_edges(self) -> int:
        """The number of edges in the graph."""
        return self.neighbour_index.shape[1]

    def to(self, device: torch.device | str) -> AtomicGraph:
        """Create a copy of this graph on the specified device."""

        def process_dict(d):
            return {key: value.to(device) for key, value in d.items()}

        return AtomicGraph(
            Z=self.Z.to(device),
            positions=self._positions.to(device),
            neighbour_index=self.neighbour_index.to(device),
            cell=self.cell.to(device),
            neighbour_offsets=self.neighbour_offsets.to(device),
            atom_labels=process_dict(self.atom_labels),
            edge_labels=process_dict(self.edge_labels),
            structure_labels=process_dict(self.structure_labels),
        )

    def __getitem__(self, key: str) -> Tensor:
        """
        Get the labels for the specified key.

        Parameters
        ----------
        key
            The name of the label to retrieve.
        """
        for labels in (
            self.atom_labels,
            self.edge_labels,
            self.structure_labels,
        ):
            if key in labels:
                return labels[key]
        raise KeyError(f"Could not find label '{key}'")

    def __contains__(self, key: str) -> bool:
        return (
            key in self.atom_labels
            or key in self.edge_labels
            or key in self.structure_labels
        )

    def is_local_property(self, x: Tensor) -> bool:
        """
        Check whether a tensor is a local property, based on its shape.

        Parameters
        ----------
        x
            The tensor to check.
        """
        return len(x.shape) > 0 and (x.shape[0] == self.n_atoms)

    def __repr__(self) -> str:
        device = self.Z.device
        _dict = dict(
            Z=self.Z,
            positions=self._positions,
            neighbour_index=self.neighbour_index,
            **self.atom_labels,
            **self.edge_labels,
            **self.structure_labels,
        )
        return f"AtomicGraph({shape_repr(_dict, sep=', ')}, device={device})"


def convert_to_atomic_graph(
    structure: ase.Atoms,
    cutoff: float,
    atom_labels: list[str] | None = None,
    structure_labels: list[str] | None = None,
) -> AtomicGraph:
    """
    Convert an ASE Atoms object to an AtomicGraph.

    Parameters
    ----------
    structures
        The ASE Atoms object.
    cutoff
        The cutoff distance for neighbour finding.
    labels
        The names of any additional labels to include in the graph.
        These must be present in either the `atoms.info`
        or `atoms.arrays` dict. If not provided, all possible labels
        will be included.
    """

    atom_info, structure_info = extract_information(
        structure, atom_labels, structure_labels
    )
    i, j, offsets = neighbor_list("ijS", structure, cutoff)
    return AtomicGraph(
        Z=torch.LongTensor(structure.numbers),
        positions=torch.FloatTensor(structure.positions),
        neighbour_index=torch.LongTensor(np.vstack([i, j])),
        cell=torch.FloatTensor(structure.cell.array),
        neighbour_offsets=torch.LongTensor(offsets),
        atom_labels=atom_info,
        structure_labels=structure_info,
    )


def extract_information(
    atoms: ase.Atoms,
    atom_keys: list[str] | None = None,
    structure_keys: list[str] | None = None,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """
    Extract any additional information from the atoms object.

    Parameters
    ----------
    atoms
        The ASE Atoms object.
    atom_keys
        The names of any additional per-atom labels to include in the graph.
        If not provided, all possible labels will be included.
    structure_keys
        The names of any additional per-structure labels to include in the
        graph. If not provided, all possible labels will be included.

    Returns
    -------
    atom_info
        The per-atom labels as tensors
    structure_info
        The per-structure labels as tensors
    """

    return (
        extract_tensors(
            atoms.arrays, atom_keys, ignore=["numbers", "positions"]
        ),
        extract_tensors(atoms.info, structure_keys, ignore=["cell", "pbc"]),
    )


def extract_tensors(
    info_dict: dict[str, object],
    labels: list[str] | None,
    ignore: list[str] | None = None,
) -> dict[str, Tensor]:
    """
    Grab any tensors from the info dict.

    If labels is None, all items will be included. We warn if any of the
    items are not tensors, but continue anyway.
    If labels are provided, they must be present in the info dict,
    and be able to be converted to tensors.
    """

    strict = labels is not None
    ignore = ignore or []

    if strict:
        missing = set(labels) - set(info_dict.keys())
        if missing:
            raise KeyError(
                f"The following labels are not present in the info dict: "
                f"{', '.join(missing)}"
            )

    labels = labels if labels is not None else list(info_dict.keys())
    tensor_dict = {}

    for key, value in info_dict.items():
        if key not in labels or key in ignore:
            continue

        maybe_tensor = as_possible_tensor(value)

        if maybe_tensor is None:
            if strict:
                raise ValueError(
                    f"Label '{key}' is not a tensor and cannot be included."
                )
        else:
            if maybe_tensor.dtype == torch.float64:
                maybe_tensor = maybe_tensor.float()
            tensor_dict[key] = maybe_tensor

    return tensor_dict


def convert_to_atomic_graphs(
    structures: Iterable[ase.Atoms] | ase.Atoms,
    cutoff: float,
    atom_labels: list[str] | None = None,
    structure_labels: list[str] | None = None,
) -> list[AtomicGraph]:
    """
    Convert a collection of ASE Atoms into a list of AtomicGraphs.

    Parameters
    ----------
    structures
        The ASE Atoms objects.
    cutoff
        The cutoff distance for neighbour finding.
    labels
        The names of any additional labels to include in the graph.
        These must be present in either the `atoms.info`
        or `atoms.arrays` dict. If not provided, all possible labels
        will be included.
    """
    if atom_labels is None and structure_labels:
        atom_labels = []
    if structure_labels is None and atom_labels:
        structure_labels = []
    if isinstance(structures, ase.Atoms):
        structures = [structures]
    return [
        convert_to_atomic_graph(s, cutoff, atom_labels, structure_labels)
        for s in structures
    ]



import numpy as np
import pytest
import torch
from graph_pes.util import (
    as_possible_tensor,
    differentiate,
    pairs,
    require_grad,
    shape_repr,
    to_chem_symbol,
)


def test_pairs():
    assert list(pairs([1, 2, 3])) == [(1, 2), (2, 3)]


def test_shape_repr():
    d = dict(
        a=torch.rand(1, 2),
        b=torch.rand(1),
    )
    assert shape_repr(d) == "a=[1,2], b=[1]"


def test_to_chem_symbol():
    assert to_chem_symbol(1) == "H"
    assert to_chem_symbol(6) == "C"
    assert to_chem_symbol(118) == "Og"


possible_tensors = [
    (1, True),
    (1.0, True),
    ([1, 2, 3], True),
    (torch.tensor([1, 2, 3]), True),
    (np.array([1, 2, 3]), True),
    (np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64), True),
    ("hello", False),
]


@pytest.mark.parametrize("obj, can_be_converted", possible_tensors)
def test_as_possible_tensor(obj, can_be_converted):
    if can_be_converted:
        assert isinstance(as_possible_tensor(obj), torch.Tensor)
    else:
        assert as_possible_tensor(obj) is None


def test_require_grad():
    # first, check that sensible error messages are displayed if attempting
    # to use this in a no_grad context:
    with torch.no_grad(), pytest.raises(
        RuntimeError, match="calculate gradients"
    ):
        x = torch.zeros(1)
        with require_grad(x):
            pass

    # now check that this thing actually works:
    y = torch.zeros(1, requires_grad=False)
    with require_grad(y):
        assert y.requires_grad
    assert not y.requires_grad

    z = torch.zeros(1, requires_grad=True)
    with require_grad(z):
        assert z.requires_grad
    assert z.requires_grad


def test_get_gradient():
    # test nice error message
    x = torch.tensor([1.0, 2.0, 3.0])
    y = x.sum()
    with pytest.raises(ValueError, match="must be the result of a computation"):
        dy_dx = differentiate(y, x)

    # test that it works
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = x.sum()
    dy_dx = differentiate(y, x)
    assert torch.allclose(dy_dx, torch.ones_like(x))

    # test that it works with a non-scalar y
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = x**2
    dy_dx = differentiate(y, x)
    assert torch.allclose(dy_dx, 2 * x)

    # test that it works if x is not part of the computation graph
    x = torch.tensor([1.0, 2.0, 3.0])
    z = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    y = z.sum()
    dy_dx = differentiate(y, x)
    assert torch.allclose(dy_dx, torch.zeros_like(x))

    # finally, we want to test that the gradient itself has a gradient
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = (x**2).sum()
    dy_dx = differentiate(y, x)
    dy_dx2 = differentiate(dy_dx, x)
    assert torch.allclose(dy_dx2, 2 * torch.ones_like(x))



from __future__ import annotations

from contextlib import contextmanager
from functools import wraps

import matplotlib.pyplot as plt
import numpy as np
import torch
from ase import Atoms
from matplotlib.ticker import MaxNLocator

from .core import GraphPESModel, get_predictions
from .data.atomic_graph import AtomicGraph, convert_to_atomic_graphs
from .data.batching import AtomicGraphBatch
from .transform import Identity, Transform
from .util import Keys

_my_style = {
    "figure.figsize": (3, 3),
    "axes.spines.right": False,
    "axes.spines.top": False,
}


@contextmanager
def nice_style():
    """
    Context manager to use my home-made plt style within a context
    """
    with plt.rc_context(_my_style):
        yield


def my_style(func):
    """
    Decorator to use my home-made plt style within a function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        with plt.rc_context(_my_style):
            return func(*args, **kwargs)

    return wrapper


def move_axes(ax):
    """
    Move the axes to the center of the figure
    """
    ax.spines["left"].set_position(("outward", 10))
    ax.spines["bottom"].set_position(("outward", 10))


@my_style
def parity_plot(
    model: GraphPESModel,
    graphs: AtomicGraphBatch | list[AtomicGraph],
    property: Keys,
    property_label: str | None = None,
    transform: Transform | None = None,
    units: str | None = None,
    ax: plt.Axes | None = None,  # type: ignore
    **scatter_kwargs,
):
    r"""
    A nicely formatted parity plot of model predictions vs ground truth
    for the given :code:`property`.

    Parameters
    ----------
    model
        The model to for generating predictions.
    graphs
        The graphs to make predictions on.
    property
        The property to plot, e.g. :code:`Keys.ENERGY`.
    property_label
        The string that the property is indexed by on the graphs. If not
        provided, defaults to the value of :code:`property`, e.g.
        :code:`Keys.ENERGY` :math:`\rightarrow` :code:`"energy"`.
    transform
        The transform to apply to the predictions and labels before plotting.
        If not provided, no transform is applied.
    units
        The units of the property, for labelling the axes. If not provided, no
        units are used.
    ax
        The axes to plot on. If not provided, the current axes are used.
    scatter_kwargs
        Keyword arguments to pass to :code:`plt.scatter`.

    Examples
    --------
    Default settings (no units, transforms or custom scatter keywords):

    .. code-block:: python

        parity_plot(model, train, Keys.ENERGY)

    .. image:: notebooks/Cu-LJ-default-parity.svg
        :align: center

    Custom settings, as seen in
    :doc:`this example notebook <notebooks/example>`:

    .. code-block:: python

        from graph_pes.transform import DividePerAtom
        from graph_pes.util import Keys

        parity_plot(
            model,
            train,
            Keys.ENERGY,
            transform=DividePerAtom(),
            units="eV/atom",
            c="royalblue",
            label="Train",
        )

        ...

    .. image:: notebooks/Cu-LJ-parity.svg
        :align: center
    """
    # deal with defaults
    transform = transform or Identity()
    if property_label is None:
        property_label = property.value

    # get the predictions and labels
    if isinstance(graphs, list):
        graphs = AtomicGraphBatch.from_graphs(graphs)

    ground_truth = transform(graphs[property_label], graphs).detach()
    predictions = transform(
        get_predictions(model, graphs, {property: property_label})[
            property_label
        ],
        graphs,
    ).detach()

    # plot
    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=0, clip_on=False)
    scatter_kwargs = {**default_kwargs, **scatter_kwargs}
    ax.scatter(ground_truth, predictions, **scatter_kwargs)

    # get a point guaranteed to be on the plot
    z = ground_truth.view(-1)[0].item()
    ax.axline((z, z), slope=1, c="k", ls="--", lw=1)

    # aesthetics
    axis_label = f"{property_label}  ({units})" if units else property_label
    ax.set_xlabel(f"True {axis_label}")
    ax.set_ylabel(f"Predicted {axis_label}")
    ax.set_aspect("equal", "datalim")
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()
    ax.set_xlim(min(x0, y0), max(x1, y1))
    ax.set_ylim(min(x0, y0), max(x1, y1))
    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))


@my_style
def dimer_curve(
    model: GraphPESModel,
    system: str,
    units: str | None = None,
    set_to_zero: bool = True,
    rmin: float = 0.9,
    rmax: float = 5.0,
    ax: plt.Axes | None = None,  # type: ignore
    **plot_kwargs,
):
    r"""
    A nicely formatted dimer curve plot for the given :code:`system`.

    Parameters
    ----------
    model
        The model to for generating predictions.
    system
        The dimer system. Should be one of: a single element, e.g. :code:`"Cu"`,
        or a pair of elements, e.g. :code:`"CuO"`.
    units
        The units of the energy, for labelling the axes. If not provided, no
        units are used.
    set_to_zero
        Whether to set the energy of the dimer at :code:`rmax` to be zero.
    rmin
        The minimum seperation to consider.
    rmax
        The maximum seperation to consider.
    ax
        The axes to plot on. If not provided, the current axes are used.
    plot_kwargs
        Keyword arguments to pass to :code:`plt.plot`.

    Examples
    --------
    See :doc:`this example notebook <notebooks/example>`:

    .. code-block:: python

        dimer_curve(model, "Cu", units="eV", label="Final", c="C1")

    .. image:: notebooks/Cu-LJ-dimer.svg
        :align: center
    """

    trial_atoms = Atoms(system)
    if len(trial_atoms) != 2:
        system = system + "2"

    rs = np.linspace(rmin, rmax, 200)
    atoms = [Atoms(system, positions=[[0, 0, 0], [r, 0, 0]]) for r in rs]
    graphs = convert_to_atomic_graphs(atoms, cutoff=rmax + 0.1)
    batch = AtomicGraphBatch.from_graphs(graphs)

    with torch.no_grad():
        energy = model(batch).numpy()
    if set_to_zero:
        energy -= energy[-1]

    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=1, c="k")
    plot_kwargs = {**default_kwargs, **plot_kwargs}
    ax.plot(rs, energy, **plot_kwargs)

    limiting_energy = energy[-1]
    if (energy[:-1] < limiting_energy).any():
        well_depth = limiting_energy - energy[:-1].min()
    else:
        well_depth = 0.1
    bottom = limiting_energy - well_depth * 1.1
    top = limiting_energy + well_depth * 1.1
    ax.set_ylim(bottom, top)
    ax.set_xlabel("r (Å)")
    ax.set_ylabel(f"Dimer Energy ({units})" if units else "Dimer Energy")

    first_in_view = np.where(energy < top)[0][0]
    ax.set_xlim(rs[first_in_view].item() - 0.2, rs[-1] + 0.2)
    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))



import pytest
import torch
from ase import Atoms
from graph_pes.core import get_predictions
from graph_pes.data import AtomicGraphBatch, convert_to_atomic_graph
from graph_pes.models.pairwise import LennardJones
from graph_pes.util import Keys

no_pbc = convert_to_atomic_graph(
    Atoms("H2", positions=[(0, 0, 0), (0, 0, 1)], pbc=False),
    cutoff=1.5,
)
pbc = convert_to_atomic_graph(
    Atoms("H2", positions=[(0, 0, 0), (0, 0, 1)], pbc=True, cell=(2, 2, 2)),
    cutoff=1.5,
)


def test_predictions():
    expected_shapes = {
        Keys.ENERGY: (),
        Keys.FORCES: (2, 3),
        Keys.STRESS: (3, 3),
    }

    model = LennardJones()

    # by default, get_predictions returns energy and forces on
    # structures with no cell:
    predictions = get_predictions(model, no_pbc)
    assert set(predictions.keys()) == {"energy", "forces"}

    for key in Keys.ENERGY, Keys.FORCES:
        assert predictions[key.value].shape == expected_shapes[key]

    # if we ask for stress, we get an error:
    with pytest.raises(ValueError):
        get_predictions(model, no_pbc, {Keys.STRESS: "stress"})

    # with pbc structures, we should get all three predictions
    predictions = get_predictions(model, pbc)
    assert set(predictions.keys()) == {"energy", "forces", "stress"}

    for key in Keys.ENERGY, Keys.FORCES, Keys.STRESS:
        assert predictions[key.value].shape == expected_shapes[key]

    # check that requesting a subset of predictions works, and that
    # the names are correctly mapped:
    predictions = get_predictions(model, no_pbc, {Keys.ENERGY: "total_energy"})
    assert set(predictions.keys()) == {"total_energy"}
    assert predictions["total_energy"].shape == expected_shapes[Keys.ENERGY]


def test_batched_prediction():
    batch = AtomicGraphBatch.from_graphs([pbc, pbc])

    expected_shapes = {
        Keys.ENERGY: (2,),  # two structures
        Keys.FORCES: (4, 3),  # four atoms
        Keys.STRESS: (2, 3, 3),  # two structures
    }

    predictions = get_predictions(LennardJones(), batch)

    for key in Keys.ENERGY, Keys.FORCES, Keys.STRESS:
        assert predictions[key.value].shape == expected_shapes[key]


def test_isolated_atom():
    atom = Atoms("H", positions=[(0, 0, 0)], pbc=False)
    graph = convert_to_atomic_graph(atom, cutoff=1.5)
    assert graph.n_edges == 0

    predictions = get_predictions(LennardJones(), graph)
    assert torch.allclose(predictions["forces"], torch.zeros(1, 3))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable

import torch
import torch.nn as nn
from graph_pes.util import MAX_Z, pairs
from torch import Tensor

# TODO support access to .data via property and setter on ConstrainedParameter
# / cleanup the ConstrainedParameter class


class MLP(nn.Module):
    """
    A multi-layer perceptron model, alternating linear layers and activations.

    Parameters
    ----------
    layers: List[int]
        The number of nodes in each layer.
    activation: str or nn.Module
        The activation function to use: either a named activation function
        from `torch.nn`, or a `torch.nn.Module` instance.
    activate_last: bool
        Whether to apply the activation function after the last linear layer.
    bias: bool
        Whether to include bias terms in the linear layers.

    Examples
    --------
    >>> import torch
    >>> from graph_pes.nn import MLP
    >>> model = MLP([10, 5, 1])
    >>> model
    MLP(10 → 5 → 1, activation=CELU())
    >>> MLP([10, 5, 1], activation=torch.nn.ReLU())
    MLP(10 → 5 → 1, activation=ReLU())
    >>> MLP([10, 5, 1], activation="Tanh")
    MLP(10 → 5 → 1, activation=Tanh())
    """

    def __init__(
        self,
        layers: list[int],
        activation: str | nn.Module = "CELU",
        activate_last: bool = False,
        bias: bool = True,
    ):
        super().__init__()

        self.activation = (
            parse_activation(activation)
            if isinstance(activation, str)
            else activation
        )
        self.activate_last = activate_last

        self.linear_layers = nn.ModuleList(
            [nn.Linear(_in, _out, bias=bias) for _in, _out in pairs(layers)]
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Perform a forward pass through the network.

        Parameters
        ----------
        x
            The input to the network.
        """
        for i, linear in enumerate(self.linear_layers):
            x = linear(x)
            last_layer = i == len(self.linear_layers) - 1
            if not last_layer or self.activate_last:
                x = self.activation(x)

        return x

    @property
    def input_size(self):
        """The size of the input to the network."""
        return self.linear_layers[0].in_features

    @property
    def output_size(self):
        """The size of the output of the network."""
        return self.linear_layers[-1].out_features

    @property
    def layer_widths(self):
        """The widths of the layers in the network."""
        inputs = [layer.in_features for layer in self.linear_layers]
        return inputs + [self.output_size]

    def __repr__(self):
        layers = " → ".join(map(str, self.layer_widths))
        return f"MLP({layers}, activation={self.activation})"


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.softplus(x) - self.shift


def parse_activation(act: str) -> torch.nn.Module:
    """
    Parse a string into a PyTorch activation function.

    Parameters
    ----------
    act
        The activation function to parse.

    Returns
    -------
    torch.nn.Module
        The parsed activation function.
    """
    activation = getattr(torch.nn, act, None)
    if activation is None:
        raise ValueError(f"Activation function {act} not found in `torch.nn`.")
    return activation()


class Product(nn.Module):
    def __init__(self, components: list[nn.Module]):
        super().__init__()
        self.components = nn.ModuleList(components)

    def forward(self, x):
        out = 1
        for component in self.components:
            out = out * component(x)
        return out


class PerSpeciesParameter(torch.nn.Parameter):
    """
    A parameter that is indexed by unqiue atomic numbers.

    Lazily keeps track of which atomic numbers have been accessed
    so as to give accurate counts of trainable parameters.

    Instantiate using the `PerSpeciesParameter.of_dim` class method.

    Examples
    --------
    >>> import torch
    >>> from graph_pes.nn import PerSpeciesParameter
    >>> param = PerSpeciesParameter.of_dim(10)
    >>> # do some computation involving some atomic numbers
    >>> ...
    >>> param[1], param[6]  # access the parameters for H and C at some point
    >>> param.numel()  # get the number of trainable parameters
    20
    >>> param
    PerSpeciesParameter({
         Z : (10,)
         3 : [ 0.6450, -0.1432, -0.5998,  ...,  0.6023, -0.3033,  0.3528],
        12 : [-0.2076, -1.3792, -0.0791,  ..., -1.5645,  1.5325, -0.7080]
    }, requires_grad=True)
    """

    # include these args for type hint reasons, class construction
    # actually takes place in __new__, as defined in PyTorch
    def __init__(self, data: Tensor, requires_grad: bool = True):
        super().__init__()

        # we lazily keep track of which atomic numbers have been accessed
        # so as to provide a correct number of trainable weights
        self._accessed_Zs = set()

    @classmethod
    def of_dim(
        cls,
        dim: int,
        requires_grad: bool = True,
        generator: Callable[[tuple[int, int]], Tensor]
        | int
        | float
        | None = None,
    ):
        """
        Create a `PerSpeciesParameter` of the given dimension.

        Parameters
        ----------
        dim
            The size of the parameter.
        generator
            The generator to use to create the parameter. Defaults to
            `torch.randn`.
        requires_grad
            Whether the parameter should be trainable.
        """
        if isinstance(generator, (int, float)):
            data = torch.full((MAX_Z, dim), generator).float()
        elif generator is None:
            data = torch.randn(MAX_Z, dim)
        else:
            data = generator((MAX_Z, dim))
        return PerSpeciesParameter(data=data, requires_grad=requires_grad)

    def __getitem__(self, Z: int | Tensor) -> Tensor:
        """
        Index the values corresponding to the given atomic number/s.

        Parameters
        ----------
        Z
            The atomic number/s of the parameter to get.
        """

        if isinstance(Z, int):
            self._accessed_Zs.add(Z)
        else:
            for Z_i in torch.unique(Z):
                self._accessed_Zs.add(Z_i.item())
        return super().__getitem__(Z)

    def __setitem__(self, Z: int | Tensor, value: Tensor):
        """
        Index the values corresponding to the given atomic number/s.

        Parameters
        ----------
        Z
            The atomic number/s of the parameter to get.
        """

        if isinstance(Z, int):
            self._accessed_Zs.add(Z)
        else:
            for Z_i in torch.unique(Z):
                self._accessed_Zs.add(Z_i.item())
        return super().__setitem__(Z, value)

    def numel(self) -> int:
        """Get the number of trainable parameters."""

        return sum(self[Z].numel() for Z in self._accessed_Zs)

    def __repr__(self) -> str:
        torch.set_printoptions(threshold=3)
        Zs = sorted(self._accessed_Zs)
        with torch.no_grad():
            matrix = str(self.data[Zs])
        matrix = matrix[8:-2]  # remove "tensor([" and "])"

        # generate a dictionary of atomic numbers to
        # their (nicely formatted) values
        lines = []
        for z, values in zip(Zs, matrix.split("\n")):
            lines.append(f"{z:3d} : {values.strip()}")

        lines = "\n    ".join(lines)

        torch.set_printoptions(profile="default")

        return f"""\
PerSpeciesParameter({{
      Z : {tuple(self.shape[1:])}
    {lines}
}}, requires_grad={self.requires_grad})"""


class PerSpeciesEmbedding(torch.nn.Module):
    """
    A per-speices equivalent of `torch.nn.Embedding`.

    Parameters
    ----------
    dim
        The dimension of the embedding.

    Examples
    --------
    >>> embedding = PerSpeciesEmbedding(10)
    >>> embedding(graph.Z)  # graph.Z is a tensor of atomic numbers
    <tensor of shape (graph.n_atoms, 10)>
    """

    def __init__(self, dim: int):
        super().__init__()
        self._embeddings = PerSpeciesParameter.of_dim(dim)

    def forward(self, Z: Tensor) -> Tensor:
        return self._embeddings[Z]


class ConstrainedParameter(nn.Module, ABC):
    """
    Abstract base class for constrained parameters.

    Implementations should override the `_constrained_value` property.

    Parameters
    ----------
    x
        The initial value of the parameter.
    requires_grad
        Whether the parameter should be trainable.
    """

    def __init__(self, x: torch.Tensor, requires_grad: bool = True):
        super().__init__()
        self._parameter = nn.Parameter(x, requires_grad)

    @property
    @abstractmethod
    def constrained_value(self) -> torch.Tensor:
        """generate the constrained value"""

    def _do_math(self, other, function, rev=False):
        if isinstance(other, ConstrainedParameter):
            other_value = other.constrained_value
        else:
            other_value = other

        if rev:
            return function(other_value, self.constrained_value)
        else:
            return function(self.constrained_value, other_value)

    def __add__(self, other):
        return self._do_math(other, torch.add)

    def __radd__(self, other):
        return self._do_math(other, torch.add, rev=True)

    def __sub__(self, other):
        return self._do_math(other, torch.sub)

    def __rsub__(self, other):
        return self._do_math(other, torch.sub, rev=True)

    def __mul__(self, other):
        return self._do_math(other, torch.mul)

    def __rmul__(self, other):
        return self._do_math(other, torch.mul, rev=True)

    def __truediv__(self, other):
        return self._do_math(other, torch.true_divide)

    def __rtruediv__(self, other):
        return self._do_math(other, torch.true_divide, rev=True)

    def __pow__(self, other):
        return self._do_math(other, torch.pow)

    def __rpow__(self, other):
        return self._do_math(other, torch.pow, rev=True)

    def log(self):
        return torch.log(self.constrained_value)

    def sqrt(self):
        return torch.sqrt(self.constrained_value)

    def __repr__(self):
        t = self.constrained_value
        if t.numel() == 1:
            return f"{self.__class__.__name__}({t.item():.4f})"
        return f"{self.__class__.__name__}({t})"

    def __neg__(self):
        return self._do_math(0, torch.sub, rev=True)


class PositiveParameter(ConstrainedParameter):
    """
    Drop-in replacement for :class:`torch.nn.Parameter`. An internal
    exponentiation ensures that the parameter is always positive.

    Parameters
    ----------
    x
        The initial value of the parameter. Must be positive.
    requires_grad
        Whether the parameter should be trainable.
    """

    def __init__(self, x: torch.Tensor | float, requires_grad: bool = True):
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x)
        super().__init__(torch.log(x), requires_grad)

    @property
    def constrained_value(self):
        return torch.exp(self._parameter)



import numpy as np
import torch
from ase import Atoms
from graph_pes.data import convert_to_atomic_graph, convert_to_atomic_graphs
from graph_pes.data.batching import AtomicGraphBatch
from graph_pes.transform import Identity, PerAtomScale, PerAtomShift

structure = Atoms("H2", positions=[(0, 0, 0), (0, 0, 1)])
structure.info["energy"] = -1.0
structure.arrays["forces"] = torch.zeros((2, 3))
graph = convert_to_atomic_graph(structure, cutoff=1.5)


def test_identity():
    transform = Identity()
    x = torch.arange(10).float()

    assert transform.forward(x, graph).equal(x)
    assert transform.inverse(x, graph).equal(x)


def test_is_local_property():
    energy = graph["energy"]
    assert not graph.is_local_property(energy)

    forces = graph["forces"]
    assert graph.is_local_property(forces)


def test_per_atom_transforms():
    # create some fake structures to test shift and scale fitting

    structures = []

    # (num H atoms, num C atoms)
    nums = [(3, 0), (4, 8), (5, 2)]
    H_energy, C_energy = -4.5, -10.0

    for n_H, n_C in nums:
        atoms = Atoms("H" * n_H + "C" * n_C)
        atoms.info["energy"] = n_H * H_energy + n_C * C_energy
        local_prop = [H_energy] * n_H + [C_energy] * n_C
        atoms.arrays["local_prop"] = local_prop
        atoms.arrays["forces"] = np.zeros((n_H + n_C, 3))
        structures.append(atoms)

    graphs = convert_to_atomic_graphs(structures, cutoff=1.5)
    batch = AtomicGraphBatch.from_graphs(graphs)

    # fit shift to the total energies
    shift = PerAtomShift(trainable=False)
    total_energies = batch["energy"]
    shift.fit(total_energies, batch)
    shifted_total_energies = shift(total_energies, batch)

    # shape preservation
    assert shifted_total_energies.shape == total_energies.shape
    # learn the correct shifts
    assert torch.allclose(
        shift.shift[torch.tensor([1, 6])].detach().squeeze(),
        torch.tensor([H_energy, C_energy]),
    )

    centered_energy = shift(total_energies, batch)
    assert torch.allclose(
        centered_energy,
        torch.zeros_like(centered_energy),
        atol=1e-5,
    )
    assert not centered_energy.requires_grad

    # fit shift to the local energies
    shift = PerAtomShift(trainable=False)
    local_energies = batch["local_prop"]

    shift.fit(local_energies, batch)
    shifted_local_energies = shift(local_energies, batch)
    assert shifted_local_energies.shape == local_energies.shape
    assert torch.allclose(
        shift.shift[torch.tensor([1, 6])].detach().squeeze(),
        torch.tensor([H_energy, C_energy]),
    )

    centered_local_energy = shift(local_energies, batch)
    assert torch.allclose(
        centered_local_energy,
        torch.zeros_like(centered_local_energy),
        atol=1e-5,
    )
    assert not centered_local_energy.requires_grad

    # test scaling forces
    scale = PerAtomScale(trainable=False)
    forces = batch["forces"]
    scale.fit(forces, batch)

    scaled_forces = scale(forces, batch)
    assert scaled_forces.shape == forces.shape



import torch
from graph_pes.nn import (
    MLP,
    PerSpeciesEmbedding,
    PerSpeciesParameter,
    PositiveParameter,
)
from graph_pes.util import MAX_Z


def test_per_species_parameter():
    """Test the PerSpeciesParameter class."""

    psp = PerSpeciesParameter.of_dim(5)
    assert psp.data.shape == (MAX_Z, 5)

    # nothing has been accessed yet, so there should be no
    # trainable parameters
    assert psp.numel() == 0

    # access the parameter for hydrogen
    psp[1]
    assert psp.numel() == 5

    embedding = PerSpeciesEmbedding(10)
    Z = torch.tensor([1, 2, 3, 4, 5])
    assert embedding(Z).shape == (5, 10)
    assert embedding.parameters().__next__().numel() == 50

    # test default value init
    assert PerSpeciesParameter.of_dim(1, generator=1.0).data.allclose(
        torch.ones(MAX_Z)
    )


def test_mlp():
    mlp = MLP([10, 20, 1])

    # test behaviour
    assert mlp(torch.zeros(10)).shape == (1,)

    # test properties
    assert mlp.input_size == 10
    assert mlp.output_size == 1

    # test internals
    assert len(mlp.linear_layers) == 2

    # test nice repr
    assert "MLP(10 → 20 → 1" in str(mlp)


def test_positive_parameter():
    x = torch.tensor([1, 2, 3]).float()
    positive_x = PositiveParameter(x)

    a = torch.tensor([-1, 0, 1]).float()

    assert torch.allclose(positive_x + a, x + a)
    assert torch.allclose(positive_x - a, x - a)
    assert torch.allclose(positive_x * a, x * a)
    assert torch.allclose(positive_x / a, x / a)
    assert torch.allclose(positive_x.log(), x.log())



from __future__ import annotations

from contextlib import contextmanager
from enum import Enum
from typing import Iterator, Sequence, TypeVar, overload

import torch
from ase.data import chemical_symbols
from torch import Tensor

T = TypeVar("T")


MAX_Z = 118
"""The maximum atomic number in the periodic table."""


T = TypeVar("T")


@overload
def pairs(a: Sequence[T]) -> Iterator[tuple[T, T]]:
    ...


@overload
def pairs(a: Tensor) -> Iterator[tuple[Tensor, Tensor]]:
    ...


def pairs(a) -> Iterator[tuple[T, T] | tuple[Tensor, Tensor]]:
    """
    Iterate over pairs of elements in `a`

    Parameters
    ----------
    a
        The sequence or tensor to iterate over.

    Example
    -------
    >>> list(pairs([1, 2, 3]))
    [(1, 2), (2, 3)]

    >>> list(pairs(Tensor([1, 2, 3])))
    [(1, 2), (2, 3)]
    """
    return zip(a, a[1:])


def shape_repr(dict_of_tensors: dict[str, Tensor], sep: str = ", ") -> str:
    """
    Generate a string representation of the shapes of the tensors in
    `dict_of_tensors`.

    Parameters
    ----------
    dict_of_tensors: dict[str, Tensor]
        The dictionary of tensors to represent.
    sep: str
        The separator to use between each tensor.

    Returns
    -------
    str
        The string representation of the shapes of the tensors.

    Example
    -------
    >>> shape_repr({'a': torch.zeros(3), 'b': torch.zeros(4)})
    'a=[3], b=[4]'
    """

    def _get_shape(tensor: Tensor) -> str:
        if len(tensor.shape) == 0:
            return f"{tensor.item():.2f}"
        return "[" + ",".join(str(i) for i in tensor.shape) + "]"

    return sep.join(f"{k}={_get_shape(v)}" for k, v in dict_of_tensors.items())


def to_chem_symbol(z: int):
    """
    Convert an atomic number to a chemical symbol.

    Parameters
    ----------
    z: int
        The atomic number.

    Returns
    -------
    str
        The chemical symbol.

    Example
    -------
    >>> to_chem_symbol(1)
    'H'
    >>> to_chem_symbol(118)
    'Og'
    """
    return chemical_symbols[z]


def as_possible_tensor(value: object) -> Tensor | None:
    """
    Convert a value to a tensor if possible.

    Parameters
    ----------
    value
        The value to convert.
    """

    try:
        tensor = torch.as_tensor(value)
        if tensor.dtype == torch.float64:
            tensor = tensor.float()
        return tensor

    except Exception:
        return None


class Keys(Enum):
    ENERGY = "energy"
    FORCES = "forces"
    STRESS = "stress"


def differentiate(y: torch.Tensor, x: torch.Tensor):
    if y.grad_fn is None:
        raise ValueError(
            "The tensor `y` must be the result of a computation "
            "that requires gradients. Currently, there is no "
            "gradient function associated with this tensor."
        )
    with require_grad(x):
        grad = torch.autograd.grad(
            y.sum(),
            x,
            create_graph=True,
            allow_unused=True,
        )[0]

    default = torch.zeros_like(x, requires_grad=True)
    return grad if grad is not None else default


@contextmanager
def require_grad(tensor: torch.Tensor):
    # check if in a torch.no_grad() context: if so,
    # raise an error
    if not torch.is_grad_enabled():
        raise RuntimeError(
            "Autograd is disabled, but you are trying to "
            "calculate gradients. Please wrap your code in "
            "a torch.enable_grad() context."
        )

    req_grad = tensor.requires_grad
    tensor.requires_grad_(True)
    yield
    tensor.requires_grad_(req_grad)



from __future__ import annotations

from abc import ABC, abstractmethod
from math import pi as π
from typing import Callable

import torch
from jaxtyping import Float
from torch import Tensor, nn


class DistanceExpansion(nn.Module, ABC):
    r"""
    Abstract base class for an expansion function, :math:`\phi(r) :
    [0, r_{\text{cutoff}}] \rightarrow \mathbb{R}^{n_\text{features}}`.

    Subclasses should implement :meth:`expand`, which must also work over
    batches: :math:`\phi(r) : [0, r_{\text{cutoff}}]^{n_\text{batch} \times 1}
    \rightarrow \mathbb{R}^{n_\text{batch} \times n_\text{features}}`.

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__()
        self.n_features = n_features
        self.cutoff = cutoff
        self.trainable = trainable

    @abstractmethod
    def expand(
        self, r: Float[Tensor, "... 1"]
    ) -> Float[Tensor, "... n_features"]:
        r"""
        Perform the expansion.

        Parameters
        ----------
        r : torch.Tensor
            The distances to expand. Guaranteed to have shape :math:`(..., 1)`.
        """

    def forward(
        self, r: Float[Tensor, "..."]
    ) -> Float[Tensor, "... n_features"]:
        if r.shape[-1] != 1:
            r = r.unsqueeze(-1)
        return self.expand(r)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(n_features={self.n_features}, "
            f"cutoff={self.cutoff}, trainable={self.trainable})"
        )


class Bessel(DistanceExpansion):
    r"""
    The Bessel expansion:

    .. math::
        \phi_{n}(r) = \sqrt{\frac{2}{r_{\text{cut}}}}
        \frac{\sin(n \pi \frac{r}{r_\text{cut}})}{r} \quad n
        \in [1, n_\text{features}]

    where :math:`r_\text{cut}` is the cutoff radius and :math:`n` is the order
    of the Bessel function, as introduced in `Directional Message Passing for
    Molecular Graphs <http://arxiv.org/abs/2003.03123>`_.

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    frequencies
        :math:`n`, the frequencies of the Bessel functions.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__(n_features, cutoff)
        self.frequencies = nn.Parameter(
            torch.arange(1, n_features + 1) * π / cutoff,
            requires_grad=trainable,
        )
        self.pre_factor = torch.sqrt(torch.tensor(2 / cutoff))

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        numerator = self.pre_factor * torch.sin(r * self.frequencies)
        # we avoid dividing by zero by replacing any zero elements with 1
        denominator = torch.where(r != 0, torch.tensor(1.0), r)
        return numerator / denominator


class GaussianSmearing(DistanceExpansion):
    r"""
    A Gaussian smearing expansion:

    .. math::
        \phi_{n}(r) = \exp\left(-\frac{(r - \mu_n)^2}{2\sigma^2}\right)
        \quad n \in [1, n_\text{features}]

    where :math:`\mu_n` is the center of the :math:`n`'th Gaussian
    and :math:`\sigma` is a width shared across all the Gaussians.

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    centers
        :math:`\mu_n`, the centers of the Gaussians.
    coef
        :math:`\frac{1}{2\sigma^2}`, the coefficient of the exponent.
    """

    def __init__(
        self,
        n_features: int,
        cutoff: float,
        trainable: bool = True,
    ):
        super().__init__(n_features, cutoff, trainable)

        sigma = cutoff / n_features
        self.coef = nn.Parameter(
            torch.tensor(-1 / (2 * sigma**2)),
            requires_grad=trainable,
        )
        self.centers = nn.Parameter(
            torch.linspace(0, cutoff, n_features),
            requires_grad=trainable,
        )

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        offsets = r - self.centers
        return torch.exp(self.coef * offsets**2)


Envelope = Callable[[torch.Tensor], torch.Tensor]


class PolynomialEnvelope(nn.Module):
    r"""
    A thrice differentiable envelope function.

    .. math::
        E_p(r) = 1 - \frac{(p+1)(p+2)}{2}\cdot r^p + p(p+2) \cdot r^{p+1}
        - \frac{p(p+1)}{2}\cdot d^{p+2}

    where :math:`r_\text{cut}` is the cutoff radius,
    and :math:`p \in \mathbb{N}`.

    Parameters
    ----------
    cutoff : float
        The cutoff radius.
    p: int
        The order of the envelope function.
    """

    def __init__(self, cutoff: float, p: int = 6):
        super().__init__()
        self.cutoff = cutoff
        self.p = p
        self.register_buffer(
            "coefficients",
            torch.tensor(
                [
                    -(p + 1) * (p + 2) / 2,
                    p * (p + 2),
                    -(p * (p + 1)) / 2,
                ]
            ),
        )
        self.register_buffer("powers", torch.arange(p, p + 3))

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        powers = (r.unsqueeze(-1) / self.cutoff) ** self.powers
        envelope = 1 + (powers * self.coefficients).sum(dim=-1)
        return torch.where(r <= self.cutoff, envelope, torch.tensor(0.0))

    def __repr__(self):
        return f"PolynomialEnvelope(cutoff={self.cutoff}, p={self.p})"


class ExtendedPolynomialEnvelope(PolynomialEnvelope):
    def __init__(self, cutoff: float, onset: float | None = None, p: int = 6):
        if onset is None:
            onset = cutoff / 3
        super().__init__(cutoff - onset, p)
        self.onset = onset

    def __call__(self, r):
        return torch.where(
            r < self.onset,
            torch.ones_like(r),
            super().__call__(r - self.onset),
        )

    def __repr__(self):
        return (
            "ExtendedPolynomialEnvelope("
            f"cutoff={self.cutoff}, onset={self.onset}, p={self.p})"
        )



from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from graph_pes.data import AtomicGraph
from graph_pes.data.batching import AtomicGraphBatch, sum_per_structure
from graph_pes.transform import (
    Chain,
    Identity,
    PerAtomScale,
    Transform,
)
from graph_pes.util import Keys, differentiate, require_grad
from jaxtyping import Float
from torch import Tensor, nn


class GraphPESModel(nn.Module, ABC):
    r"""
    An abstract base class for all graph-based, energy-conserving models of the
    PES that make predictions of the total energy of a structure as a sum
    of local contributions:

    .. math::
        E(\mathcal{G}) = \sum_i \varepsilon_i

    To create such a model, implement :meth:`predict_local_energies`,
    which takes an :class:`AtomicGraph`, or an :class:`AtomicGraphBatch`,
    and returns a per-atom prediction of the local energy. For a simple example,
    see :class:`LennardJones <graph_pes.models.pairwise.LennardJones>`.

    Under the hood, :class:`GraphPESModel` contains an
    :class:`EnergySummation` module, which is responsible for
    summing over local energies to obtain the total energy/ies,
    with optional transformations of the local and total energies.
    By default, this learns a per-species, local energy offset and scale.

    .. note::
        All :class:`GraphPESModel` instances are also instances of
        :class:`torch.nn.Module`. This allows for easy optimisation
        of parameters, and automated save/load functionality.
    """

    # TODO: fix this for the case of an isolated atom, either by itself
    # or within a batch: perhaps that should go in sum_per_structure?
    # or maybe default to a local scale followed by a global peratomshift?
    @abstractmethod
    def predict_local_energies(
        self, graph: AtomicGraph | AtomicGraphBatch
    ) -> Float[Tensor, "graph.n_atoms"]:
        """
        Predict the (standardized) local energy for each atom in the graph.

        Parameters
        ----------
        graph
            The graph representation of the structure/s.
        """

    def forward(self, graph: AtomicGraph | AtomicGraphBatch):
        """
        Predict the total energy of the structure/s.

        Parameters
        ----------
        graph : AtomicGraph
            The graph representation of the structure/s.
        """

        # local predictions
        local_energies = self.predict_local_energies(graph).squeeze()

        # sum over atoms to get total energy
        return self._energy_summation(local_energies, graph)

    def __init__(self):
        super().__init__()
        self._energy_summation = EnergySummation()

    def __repr__(self):
        # modified from torch.nn.Module.__repr__
        # changes:
        # - don't print any modules that start with _

        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = self.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split("\n")
        child_lines = []
        for key, module in self._modules.items():
            if key.startswith("_"):
                continue
            mod_str = repr(module)
            mod_str = nn.modules.module._addindent(mod_str, 2)
            child_lines.append("(" + key + "): " + mod_str)
        lines = extra_lines + child_lines

        main_str = self._get_name() + "("
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"
        return main_str

    def __add__(self, other: GraphPESModel) -> Ensemble:
        """
        A convenient way to create a summation of two models.

        Examples
        --------
        >>> TwoBody() + ThreeBody()
        Ensemble([TwoBody(), ThreeBody()], aggregation=sum)
        """

        return Ensemble([self, other], mean=False)

    def pre_fit(self, graphs: AtomicGraphBatch, energy_label: str = "energy"):
        """
        Perform optional pre-processing of the training data.

        By default, this fits a :class:`graph_pes.transform.PerAtomShift`
        and :class:`graph_pes.transform.PerAtomScale` to the energies
        of the training data, such that, before training, a unit-Normal
        output by the underlying model will result in energy predictions
        that are distributed according to the training data.

        Parameters
        ----------
        graphs
            The training data.
        """
        self._energy_summation.fit_to_graphs(graphs, energy_label)


class EnergySummation(nn.Module):
    def __init__(
        self,
        local_transform: Transform | None = None,
        total_transform: Transform | None = None,
    ):
        super().__init__()

        # if both None, default to a per-species, local energy offset
        if local_transform is None and total_transform is None:
            local_transform = Chain(
                [PerAtomScale(), PerAtomScale()], trainable=True
            )
        self.local_transform: Transform = local_transform or Identity()
        self.total_transform: Transform = total_transform or Identity()

    def forward(self, local_energies: torch.Tensor, graph: AtomicGraphBatch):
        local_energies = self.local_transform.inverse(local_energies, graph)
        total_E = sum_per_structure(local_energies, graph)
        total_E = self.total_transform.inverse(total_E, graph)
        return total_E

    def fit_to_graphs(
        self,
        graphs: AtomicGraphBatch | list[AtomicGraph],
        energy_label: str = "energy",
    ):
        if not isinstance(graphs, AtomicGraphBatch):
            graphs = AtomicGraphBatch.from_graphs(graphs)

        for transform in [self.local_transform, self.total_transform]:
            transform.fit(graphs[energy_label], graphs)


class Ensemble(GraphPESModel):
    def __init__(self, models: list[GraphPESModel], mean: bool = True):
        super().__init__()
        self.models: list[GraphPESModel] = nn.ModuleList(models)  # type: ignore
        self.mean = mean

    def predict_local_energies(self, graph: AtomicGraph | AtomicGraphBatch):
        s = sum(m.predict_local_energies(graph).squeeze() for m in self.models)
        return s / len(self.models) if self.mean else s

    def __repr__(self):
        aggregation = "mean" if self.mean else "sum"
        return f"Ensemble({self.models}, aggregation={aggregation})"


# TODO: add training flag to this so that we don't create the graph needlessly
# when in eval mode
def get_predictions(
    pes: GraphPESModel,
    structure: AtomicGraph | AtomicGraphBatch | list[AtomicGraph],
    property_labels: dict[Keys, str] | None = None,
) -> dict[str, torch.Tensor]:
    """
    Evaluate the `pes` on `structure` to get the labels requested.

    Parameters
    ----------
    pes
        The PES to use.
    structure
        The atomic structure to evaluate.
    property_labels
        The names of the properties to return. If None, all available
        properties are returned.

    Returns
    -------
    dict[str, torch.Tensor]
        The requested properties.

    Examples
    --------
    >>> # TODO

    """

    if isinstance(structure, list):
        structure = AtomicGraphBatch.from_graphs(structure)

    if property_labels is None:
        property_labels = {
            Keys.ENERGY: "energy",
            Keys.FORCES: "forces",
        }
        if structure.has_cell:
            property_labels[Keys.STRESS] = "stress"

    else:
        if Keys.STRESS in property_labels and not structure.has_cell:
            raise ValueError("Can't predict stress without cell information.")

    predictions = {}

    # setup for calculating stress:
    if Keys.STRESS in property_labels:
        # The virial stress tensor is the gradient of the total energy wrt
        # an infinitesimal change in the cell parameters.
        # We therefore add this change to the cell, such that
        # we can calculate the gradient wrt later if required.
        #
        # See <> TODO: find reference
        actual_cell = structure.cell
        change_to_cell = torch.zeros_like(actual_cell, requires_grad=True)
        symmetric_change = 0.5 * (
            change_to_cell + change_to_cell.transpose(-1, -2)
        )
        structure.cell = actual_cell + symmetric_change
    else:
        change_to_cell = torch.zeros_like(structure.cell)

    # use the autograd machinery to auto-magically calculate forces and stress
    # from the energy
    with require_grad(structure._positions), require_grad(change_to_cell):
        energy = pes(structure)

        if Keys.ENERGY in property_labels:
            predictions[property_labels[Keys.ENERGY]] = energy

        if Keys.FORCES in property_labels:
            dE_dR = differentiate(energy, structure._positions)
            predictions[property_labels[Keys.FORCES]] = -dE_dR

        if Keys.STRESS in property_labels:
            stress = differentiate(energy, change_to_cell)
            predictions[property_labels[Keys.STRESS]] = stress

    return predictions
